#include "coefs.h"

void COEFS::fill(ex &e)
{
	if (basis.size()<=0) printf("error: COEFS: no basis %d\n", basis.size());
	else  
	{
		SC sc;
		sc.init(basis.size());
		sc.coef=(ex)0;
		sc.base=(ex)1;
		fill_recursive(e, sc, 0);  // musts be expanded!!!!
	}
}

bool COEFS::fill_recursive(ex &e, SC &sc, int level)
{
	if (level>basis.size()) return false;
	if (level==basis.size()) // last entry
	{
		ex E=expand(e);
		if (E!=0) //  non-zero coefficient
		{
			sc.coef=E;
			v.push_back(sc);
			return true;
		} else
		return false;
	}
	for (int ii=0;ii<order;ii++)
	{
		// hotfix to enforce ordering
		int i = (ii==order-1) ? 0 : ii+1;
		if (sc.order()+i<order)
		{
			ex c = coeff(e, basis[level], i);
			if (c!=0)
			{
				sc.exp[level] = i;
				ex oldbase=sc.base;
				sc.base *= pow(basis[level],i);
				fill_recursive(c, sc, level+1);
				sc.base = oldbase;
				sc.exp[level] = 0;
			}
		}
	}
	return true;
}

ex COEFS::check(ex &e)
{
	// check whether all coefficients were correctly read
	ex expr = 0;
	for (int i=0;i<v.size();i++)
	{
		expr += v[i].coef*v[i].base;
	}
	ex diff=expand(e-expr);
	if (diff!=0)
	{
		cout << "check: nonzero diff = $"<<endl << diff <<"$"<< endl<< endl;
		cout << "source: $"<<endl << e<<endl <<"$"<< endl<< endl;
		cout << "target: $"<<endl << expr<<endl <<"$"<< endl<< endl;
	}
	return diff;
}

bool COEFS::compare(SC &sc1, SC &sc2)
{
	if (sc1.N != sc2.N) return false;
	for (int i=0;i<sc1.N;i++)
		if (sc1.exp[i]!=sc2.exp[i]) return false;
	return true;
}

void COEFS::add(COEFS &CF2)
{
	for (int j=0;j<CF2.v.size();j++)
	{
		// find match for 
		int match=-1;
		for (int i=0;i<v.size();i++)
			if (compare(v[i],CF2.v[j])) {  match = j; break; }
		if (match<0)
			v.push_back(CF2.v[j]);
		else
			v[match].coef += CF2.v[j].coef;
	}
}

void COEFS::copy_basis(COEFS &CF2)
{
	for (int i=0;i<CF2.basis.size();i++) basis.push_back(CF2.basis[i]);
}
