#include "core.h"

std::string CORE::ReplaceString(std::string subject, const std::string& search, const std::string& replace) 
{
	size_t pos = 0;
	while ((pos = subject.find(search, pos)) != std::string::npos) 
	{
		subject.replace(pos, search.length(), replace);
		pos += replace.length();
	}
	return subject;
}

void CORE::coord2str(char *str, int x, int y, int z)
{
	char sx[20],sy[20],sz[20];
	if (x<0) sprintf(sx,"\\bar%d",-x); else sprintf(sx,"%d",x);
	if (y<0) sprintf(sy,"\\bar%d",-y); else sprintf(sy,"%d",y);
	if (z<0) sprintf(sz,"\\bar%d",-z); else sprintf(sz,"%d",z);

	if (DIM==1) sprintf(str,"%s",sx);
	if (DIM==2) sprintf(str,"%s%s",sx,sy);
	if (DIM==3) sprintf(str,"%s%s%s",sx,sy,sz);
}

void CORE::desc2str(char *str, int k, int j1, int j2, int j3)
{
	int j=j1+j2+j3;
	char str_i1[50], str_i2[50], str_i3[50], str_t[50];
	sprintf(str,"1");
	if (k-j>0) { if (k-j==1) sprintf(str_t,"t"); else sprintf(str_t,"t^{%d}",k-j); } else sprintf(str_t,"");
	if (j1>0) { if (j1==1) sprintf(str_i1,"x"); else sprintf(str_i1,"x^{%d}",j1); } else sprintf(str_i1,"");
	if (j2>0) { if (j2==1) sprintf(str_i2,"y"); else sprintf(str_i2,"y^{%d}",j2); } else sprintf(str_i2,"");
	if (j3>0) { if (j3==1) sprintf(str_i3,"z"); else sprintf(str_i3,"z^{%d}",j3); } else sprintf(str_i3,"");
	if (k>0) sprintf(str,"%s%s%s%s",str_t,str_i1,str_i2,str_i3);
}

void CORE::number2str(char *str, const char *s1, const char *s2, int number)
{
	if (number==0) { sprintf(str,"%s",s1); return; }
	if (number==1) { sprintf(str,"%s+%s",s1,s2); return; }
	if (number==-1) { sprintf(str,"%s-%s",s1,s2); return; }
	if (number<0) { sprintf(str,"%s%d%s",s1,number,s2); return; }
	if (number>0) { sprintf(str,"%s+%d%s",s1,number,s2); return; }
}

void CORE::desc2str(char *s_space_coeff, char* s_time_coeff, char* s_space_formula, char* s_time_formula, int l, int i1, int i2, int i3)
{
	char str_t[50], str_i1[50], str_i2[50], str_i3[50];
	number2str(str_t,"t","\\dt",-l);
	number2str(str_i1,"x","\\dl",i1);
	number2str(str_i2,"y","\\dl",i2);
	number2str(str_i3,"z","\\dl",i3);

	sprintf(s_time_coeff, "{%d}",l);
	sprintf(s_time_formula, "{%s}",str_t); 
	// fix
	sprintf(s_time_coeff, "%s",s_time_formula); 
	switch (DIM)
	{
		case 1: 
			sprintf(s_space_coeff, "{%d}",i1); 
			sprintf(s_space_formula, "{%s}",str_i1); 
			break;
		case 2: 
			sprintf(s_space_coeff, "{%d,%d}",i1,i2); 
			sprintf(s_space_formula, "{%s, %s}",str_i1,str_i2); 
			break;
		case 3: sprintf(s_space_coeff, "{%d,%d,%d}",i1,i2,i3); 
			sprintf(s_space_formula, "{%s, %s, %s}",str_i1,str_i2,str_i3); 
			break;
	}
}


void CORE::Feq_symbolic(matrix &F, int l, int x, int y, int z)
{
	for (int q=0;q<Q;q++) F(q,0) = (constvel) ? sym_feq[q] : sym_feqs[q][index(l,x,y,z)];
	// enforce sum feq[q]=1
	F(0,0)=(ex)1;
	for (int q=1;q<Q;q++) F(0,0) -= F(q,0);
}

void CORE::Meq_symbolic(matrix &F, int l, int x, int y, int z)
{
	for (int q=0;q<Q;q++) F(q,0) = (constvel) ? sym_meq[q] : sym_meqs[q][index(l,x,y,z)];
	// force 1 to zeroth order \tilde m^{eq}
	F(0,0) = (ex)1;
}

void CORE::define_equilibrium()
{
	
	if (constvel)
	{
		const_feq=matrix(Q,1);
		const_meq=matrix(Q,1);

		if (feqorder>0)
		{
			lbm.Feq(const_feq,sym_vel[0],sym_vel[1],sym_vel[2],sym_cs,Q,feqorder);
		}
		else
		{
			Feq_symbolic(const_feq,0,0,0,0);
			Meq_symbolic(const_meq,0,0,0,0);
		}
		simplify_matrix(const_feq,Q,1);

		if (feqorder>0)
		{
		}
	} else
	{
		// define equilibrium 
		feq = new matrix[Nindex];
		meq = new matrix[Nindex];
		for (int l=0;l<L;l++)
		{
			if (verbosity==1){ printf("allocating feq l %d of %d    \r", l,L-1); fflush(stdout); }
			for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
			for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
			for (int x=-(l+1);x<=l+1;x++)
			if (isin(l,x,y,z))
			{
				ex u[3];
				for (int d=0;d<3;d++) u[d]=0;
				for (int d=0;d<DIM;d++) u[d] = (constvel) ? sym_vel[d] : sym_vels[d][index(l,x,y,z)];

				feq[index(l,x,y,z)]=matrix(Q,1);
				meq[index(l,x,y,z)]=matrix(Q,1);

				if (feqorder>0)
				{
					lbm.Feq(feq[index(l,x,y,z)],u[0],u[1],u[2],sym_cs,Q,feqorder);
					simplify_matrix(feq[index(l,x,y,z)],Q,1);
				}
				else
				{
					Feq_symbolic(feq[index(l,x,y,z)],l,x,y,z);
					Meq_symbolic(meq[index(l,x,y,z)],l,x,y,z);
				}
			}
		}
		Feq_allocated=true;
	}
}

void CORE::fill_sym_vels_to_basis(COEFS &CF)
{
	// u,v,w
	for (int d=0;d<DIM;d++)
	{
		CF.basis.push_back(sym_vel[d]);
		CF.basis_tex.push_back(sym_vel[d]);
		CF.pow_dl.push_back(0);
		CF.pow_dt.push_back(0);
	}

	if (use_tay_vels)
	for (int k=1;k<=ORDER;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)	// r1,r2,r3 - powers in the multiindex r1+r2+r3=k-j
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	if (j1+j2+j3==j)
	{
		for (int d=0;d<DIM;d++)
		{
			CF.pow_dl.push_back(tay_vels[d]->get_pow_dl(k-j,j1,j2,j3));
			CF.pow_dt.push_back(tay_vels[d]->get_pow_dt(k-j,j1,j2,j3));
			
			CF.basis.push_back(tay_vels[d]->get_tay(k-j,j1,j2,j3));
			CF.basis_tex.push_back(tay_vels[d]->get_tay_tex(k-j,j1,j2,j3));
		}
	}
}

void CORE::taylor_list_vels(lst &list)
{
	for (int l=0;l<L;l++)
	for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
	for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
	for (int x=-(l+1);x<=l+1;x++)
	if (isin(l,x,y,z))
		for (int d=0;d<DIM;d++)
			list.append(sym_vels[d][index(l,x,y,z)] == tay_vels[d]->sub[index(l,x,y,z)]);
}

void CORE::taylor_list_ms(lst& list)
{
	for (int l=0;l<L;l++)
	for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
	for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
	for (int x=-(l+1);x<=l+1;x++)
	if (isin(l,x,y,z))
		for (int q=0;q<Q;q++)
			list.append(sym_ms[q][index(l,x,y,z)] == tay_ms[q]->sub[index(l,x,y,z)]);
}

void CORE::taylor_list_meqs(lst& list)
{
	for (int l=0;l<L;l++)
	for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
	for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
	for (int x=-(l+1);x<=l+1;x++)
	if (isin(l,x,y,z))
		for (int q=0;q<Q;q++)
			list.append(sym_meqs[q][index(l,x,y,z)] == tay_meqs[q]->sub[index(l,x,y,z)]);
}


void CORE::allocate()
{
	// decide what to use here
	use_tay_ms = true; // always use
	use_tay_vels = true;
	if (feqorder>0) use_tay_meqs = false;
	if (constvel) use_tay_vels = false;
	if (constvel) use_tay_meqs = false;

	// fill in symbols
	char desc[1000], s[20], desc2[20],desc3[20];
	// allocate symbolic feq
	sym_feq = new ex[Q];
	sym_f = new ex[Q];
	sym_meq = new ex[Q];
	sym_m = new ex[Q];
	for (int q=0;q<Q;q++)
	{
		sprintf(desc,"{\\tilde f}^{eq}_{%d}",q);
		sym_feq[q] = symbol(desc);
		sprintf(desc,"f_{%d}",q);
		sym_f[q] = symbol(desc);
		sprintf(desc,"{\\tilde m}^{eq}_{%d}",q);
		if (q==0)
			sym_meq[0] = (ex)1;
		else
			sym_meq[q] = symbol(desc);
		lbm.moment_symbol(desc2,q,Q);
		lbm.moment_symbol_number(desc3,q,Q);
		sprintf(desc,"%s_{%s}",desc2,desc3);
		sym_m[q] = symbol(desc);
	}
	sym_cs = symbol("{c_s}");

	sym_vel[0] = symbol("u");
	sym_vel[1] = symbol("v");
	sym_vel[2] = symbol("w");


	sym_vels = new ex*[DIM];
	for (int d=0;d<DIM;d++) sym_vels[d] = new ex[Nindex];

	if (use_tay_vels) 
	{
		tay_vels = new TAYLOR*[DIM];
		if (verbosity==1) printf("allocate tay_vels\n");
		for (int d=0;d<DIM;d++) tay_vels[d] = new TAYLOR(DIM,Q,ORDER,_index,Nindex,sym_dl,sym_dt);
		if (DIM>=1) tay_vels[0]->allocate("u", sym_vel[0]);
		if (DIM>=2) tay_vels[1]->allocate("v", sym_vel[1]);
		if (DIM>=3) tay_vels[2]->allocate("w", sym_vel[2]);
	}

	if (verbosity==1) printf("allocate sym_fs/feq/ms/meqs\n");

	sym_fs = new ex*[Q];
	sym_feqs = new ex*[Q];
	sym_ms = new ex*[Q];
	sym_meqs = new ex*[Q];

	for (int q=0;q<Q;q++)
	{
		sym_fs[q] = new ex[Nindex];
		sym_feqs[q] = new ex[Nindex];
		sym_ms[q] = new ex[Nindex];
		sym_meqs[q] = new ex[Nindex];
	}

	if (verbosity==1) printf("allocate coefs\n");
	
	coefs_efde = new ex**[Q];
	coefs_epde = new ex***[Q];
	for (int q1=0;q1<Q;q1++)
	{
		coefs_epde[q1] = new ex**[Q];
		coefs_efde[q1] = new ex*[Q];
		for (int q2=0;q2<Q;q2++)
		{
			coefs_efde[q1][q2] = new ex[Nindex];
			for (int i=0;i<Nindex;i++) coefs_efde[q1][q2][i]=0;

			coefs_epde[q1][q2] = new ex*[ORDER+1];
			for (int o=0;o<=ORDER;o++)
			{
				coefs_epde[q1][q2][o] = new ex[(ORDER+1)*(ORDER+1)*(ORDER+1)];
				for (int k1=0;k1<=ORDER;k1++) 
				for (int k2=0;k2<=ORDER;k2++) 
				for (int k3=0;k3<=ORDER;k3++) 
					coefs_epde[q1][q2][o][indO(k1,k2,k3)]=0;
			}
		}
	}

	if (verbosity==1) printf("allocate sym_vels\n");

	for (int l=0;l<L;l++)
	for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
	for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
	for (int x=-(l+1);x<=l+1;x++)
	if (isin(l,x,y,z))
	{
		coord2str(s,x,y,z);
		sprintf(desc,"u_{%d,%s}",l,s); sym_vels[0][index(l,x,y,z)] = symbol(desc);
		sprintf(desc,"v_{%d,%s}",l,s); if (DIM>=2) sym_vels[1][index(l,x,y,z)] = symbol(desc);
		sprintf(desc,"w_{%d,%s}",l,s); if (DIM>=3) sym_vels[2][index(l,x,y,z)] = symbol(desc);
		if (constvel) 
			for (int d=0;d<DIM;d++) 
				sym_vels[d][index(l,x,y,z)] = sym_vel[d];

		for (int q=0;q<Q;q++) 
		{ 
			sprintf(desc,"f_{%d,%d,%s}",q,l,s); sym_fs[q][index(l,x,y,z)] = symbol(desc); 
			sprintf(desc,"{\\tilde m^{eq}}_{%d,%d,%s}",q,l,s); sym_meqs[q][index(l,x,y,z)] = symbol(desc); 
			sprintf(desc,"{\\tilde f^{eq}}_{%d,%d,%s}",q,l,s); sym_feqs[q][index(l,x,y,z)] = symbol(desc); 

			lbm.moment_symbol(desc2,q,Q);
			lbm.moment_symbol_number(desc3,q,Q);
			if (strlen(desc3)>0)
				sprintf(desc,"%s_{%s,%d,%s}",desc2,desc3,l,s); 
			else
				sprintf(desc,"%s_{%d,%s}",desc2,l,s);
			sym_ms[q][index(l,x,y,z)] = symbol(desc);
		}
		if (l==0)
		if ((DIM==1 && x==0) || (DIM==2 && x==0 && y==0) || (DIM==3 && x==0 && y==0 && z==0))
		{
			for (int d=0;d<DIM;d++) 
				sym_vels[d][index(l,x,y,z)] = sym_vel[d];
			for (int q=0;q<Q;q++)
			{
				sym_fs[q][index(l,x,y,z)] = sym_f[q];
				sym_ms[q][index(l,x,y,z)] = sym_m[q];
			}
		}
	}

	if (verbosity==1) if (use_tay_ms) printf("allocate tay_ms\n");
	if (verbosity==1) if (use_tay_meqs) printf("allocate tay_meqs\n");

	char comp[20];
	if (use_tay_ms) tay_ms = new TAYLOR*[Q];
	if (use_tay_meqs) tay_meqs = new TAYLOR*[Q];
	for (int q=0;q<Q;q++)
	{
		if (use_tay_ms)
		{
			tay_ms[q] = new TAYLOR(DIM,Q,ORDER,_index,Nindex,sym_dl,sym_dt);
			lbm.moment_symbol(desc2,q,Q);
			lbm.moment_symbol_number(desc3,q,Q);
			if (strlen(desc3)>0)
				sprintf(comp,"%s_{%s}",desc2,desc3); 
			else
				sprintf(comp,"%s",desc2);
			tay_ms[q]->allocate(comp,sym_m[q]);
		}
		if (use_tay_meqs)
		{
			tay_meqs[q] = new TAYLOR(DIM,Q,ORDER,_index,Nindex,sym_dl,sym_dt);
			sprintf(comp,"{{\\tilde m}^{eq}_{%d}}",q); 
			tay_meqs[q]->allocate(comp,0);
		}
	}

	if (verbosity==1) printf("allocate matrices\n");

	// identity matrix
	I = matrix(Q,Q);
	empty_matrix(I);
	for (int i=0;i<Q;i++)
		I(i,i) = 1;

	// matrices E
	for (int m=0;m<Q;m++)
	{
		E[m] = matrix(Q,Q);
		for (int i=0;i<Q;i++)
		for (int j=0;j<Q;j++)
			E[m](i,j) = (i==j && j==m) ? 1 : 0;
	}

	// Matrix C that defines discrete velocities
	C = matrix(DIM,Q);
	lbm.setC(C,Q);

	for (int i=0;i<DIM;i++)
	{
		for (int j=0;j<Q;j++)
		{
			CC[i][j]=0;
			if (C(i,j)>0) CC[i][j] = 1;
			if (C(i,j)<0) CC[i][j] = -1;
			if (verbosity==1) printf("C[%d,%d]=%d ",i,j,CC[i][j]);
		}
		if (verbosity==1) printf("\n");
	}

	// set all entries to zero
	for (int q=0;q<Q;q++)
	{
		e[q] = matrix(Q,1);
		empty_matrix(e[q],Q,1);
		e[q](q,0)=(ex)1;
	}

	CP = new matrix[Nindex];
	trP = new ex[Nindex];

	// allocate matrices
	As = new matrix[Nindex];
	dBMi = new matrix[Nindex];

	allocated=true;
}

void CORE::define_matrix_M()
{
	M = matrix(Q,Q);
	Mi = matrix(Q,Q);
	lbm.setM_MRT1(M, Q);
	Mi = inverse(M);
}


void CORE::define_matrix_A_worker(matrix &_A, ex &u, ex &v, ex&w)
{
	switch (collision_model)
	{
		case SRT: lbm.setA_SRT(_A,Q,u,v,w); break;
		case MRT1: lbm.setA_MRT1(_A,Q,u,v,w); break;
		case MRT2: lbm.setA_MRT2(_A,Q,u,v,w); break;
		case CLBM1: lbm.setA_CLBM1(_A,Q,u,v,w); break;
		case CLBM2: lbm.setA_CLBM2(_A,Q,u,v,w); break;
		case MRT1_OG: lbm.setA_MRT1_OG(_A,Q,u,v,w); break;
		case MRT2_OG: lbm.setA_MRT2_OG(_A,Q,u,v,w); break;
		case CLBM1_OG: lbm.setA_CLBM1_OG(_A,Q,u,v,w); break;
		case CLBM2_OG: lbm.setA_CLBM2_OG(_A,Q,u,v,w); break;
		case GENERAL: lbm.setA_GENERAL(_A,Q,u,v,w); break;
	}
}

void CORE::define_matrix_A()
{
	// define matrix A: central matrix at [0,0,0]: A_0: A_t^x
	A = matrix(Q,Q);
	empty_matrix(A);
	define_matrix_A_worker(A,sym_vel[0],sym_vel[1],sym_vel[2]);
	simplify_matrix(A);

	if (!constvel)
	{
		for (int l=0;l<L;l++)
		{
			for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
			for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
			for (int x=-(l+1);x<=l+1;x++)
			if (isin(l,x,y,z))
			{
				As[index(l,x,y,z)] = matrix(Q,Q);
				ex u=sym_vels[0][index(l,x,y,z)]; 
				ex v=(DIM>=2) ? sym_vels[1][index(l,x,y,z)] : 0;
				ex w=(DIM>=3) ? sym_vels[2][index(l,x,y,z)] : 0;
				define_matrix_A_worker(As[index(l,x,y,z)],u,v,w);
				simplify_matrix(As[index(l,x,y,z)]);
			}
		}
	}
}

void CORE::define_matrix_B()
{
	// constant matrix B
	B = matrix(Q,Q);
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
		B(i,j) = (I.sub(A))(i,j) - (I.sub(A))(i,i);
	simplify_matrix(B);
	BMi = B.mul(Mi);
	simplify_matrix(BMi);

	if (!constvel)
	{
		for (int l=0;l<L;l++)
		{
			for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
			for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
			for (int x=-(l+1);x<=l+1;x++)
			if (isin(l,x,y,z))
			{
				matrix _A = As[index(l,x,y,z)];
				matrix _B = matrix(Q,Q);
				matrix I_A = I.sub(_A);
				empty_matrix(_B);
				for (int i=0;i<Q;i++)
				for (int j=0;j<Q;j++)
					_B(i,j) = I_A(i,j) - I_A(i,i);
				dBMi[index(l,x,y,z)] = (_B.sub(B)).mul(Mi); 
				simplify_matrix(dBMi[index(l,x,y,z)]);
			}
		}
	}
}

void CORE::compute_coefficients()
{
	// allocate P
	matrix*P = new matrix[Nindex];

	// we can use symbolic matrix B for computation
	matrix BB=B;
	lst list_BB;
	if (use_symbolic_B)
	{
		BB = matrix(Q,Q);
		ex SB = symbolic_matrix(Q,Q,"b");
		for (int i=0;i<Q;i++)
		for (int j=0;j<Q;j++)
		{
			if (i==j)
			{
				BB(i,j)=(ex)0;
			} else
			{
				BB(i,j) = SB[j+i*Q]; // symbolic
				list_BB.append(SB[j+i*Q]==B(i,j));
			}
		}
	}

	// initial setup
	for (int z=-1;(z<=1 && DIM==3) || (z<=-1);z++)
	for (int y=-1;(y<=1 && DIM>=2) || (y<=-1);y++)
	for (int x=-1;x<=1;x++)
	{
		P[index(0,x,y,z)] = matrix(Q,Q);
		empty_matrix(P[index(0,x,y,z)]);
		trP[index(0,x,y,z)] = 0;
	}
	
	
	for (int q=0;q<Q;q++)
	{
		int i=-CC[0][q];
		int j=(DIM>=2)? -CC[1][q] : 0;
		int k=(DIM>=3)? -CC[2][q] : 0;
		P[index(0,i,j,k)] = E[q].mul(BB);
	}

	// dry run
	int max=0,count=0;
	for (int l=1;l<L;l++)
	for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
	for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
	for (int x=-(l+1);x<=l+1;x++)
	if (isin(l,x,y,z))
	{
//		printf("is in l %d x %d\n",l,x);
		max++;
	}
	
	matrix sumE=matrix(Q,Q);
	for (int l=1;l<L;l++)
	{
		if (verbosity==1) printf("level %d\n",l);
		// allocate required matrices
		int nz=0; // count non-zero matrices
		// sum_z, z \in M(l+1)_x, i.e., z = x - sum c_mi (i=0..l+1) ... but z = (x,y,z) here
		for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
		for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
		for (int x=-(l+1);x<=l+1;x++)
		if (isin(l,x,y,z))
		{
			if (verbosity==1) { printf("[%d/%d] \r",count++,max); fflush(stdout); }
			P[index(l,x,y,z)] = matrix(Q,Q);
			empty_matrix(P[index(l,x,y,z)]);
			// sum_y y\in M(l)_x 
			for (int z1=-l;(z1<=l && DIM==3) || (z1<=-l);z1++)
			for (int y1=-l;(y1<=l && DIM>=2) || (y1<=-l);y1++)
			for (int x1=-l;x1<=l;x1++)
			if (isin(l-1,x1,y1,z1))
			{
				if (!P[index(l-1,x1,y1,z1)].is_zero_matrix())
				{
					empty_matrix(sumE);
					for (int q=0;q<Q;q++)
					{
						int i=-CC[0][q];
						int j=(DIM>=2)? -CC[1][q] : 0;
						int k=(DIM>=3)? -CC[2][q] : 0;
						
						// P_{x-cq} = E_q  ... P_z = E_q  where x-cq=z
						if ( 	(DIM==1 && x1+i==x) ||
							(DIM==2 && x1+i==x && y1+j==y) ||
							(DIM==3 && x1+i==x && y1+j==y && z1+k==z)
						)
						sumE = sumE.add(E[q]);
					}
					if (!sumE.is_zero_matrix())
						P[index(l,x,y,z)] = P[index(l,x,y,z)].add(P[index(l-1,x1,y1,z1)].mul(sumE));
				}
			}
			P[index(l,x,y,z)] = P[index(l,x,y,z)].mul(BB);
			trP[index(l,x,y,z)] = trace(P[index(l,x,y,z)])/(l+1);
			P[index(l,x,y,z)] = P[index(l,x,y,z)].sub( I.mul_scalar(trP[index(l,x,y,z)]) );
			simplify_matrix(P[index(l,x,y,z)]);
			if (!P[index(l,x,y,z)].is_zero_matrix()) nz++;
		}
		if (verbosity==1) printf("- non-zero P matrices: %d\n",nz);
	}

	// dry run
	if (verbosity==1) printf("compute CP:\n");
	max=0,count=0;
	for (int l=1;l<L;l++)
	for (int z=-(l+1);(z<=l+1 && DIM==3) || (z<=-(l+1));z++)
	for (int y=-(l+1);(y<=l+1 && DIM>=2) || (y<=-(l+1));y++)
	for (int x=-(l+1);x<=l+1;x++)
	if (isin(l,x,y,z))
		max++;
	
	CP[index(0,0,0,0)] = M;
	for (int l=1;l<=L;l++)
	for (int z=-l;(z<=l && DIM==3) || (z<=-l);z++)
	for (int y=-l;(y<=l && DIM>=2) || (y<=-l);y++)
	for (int x=-l;x<=l;x++)
	if (isin(l-1,x,y,z))
	{
		if (verbosity==1) { printf("[%d/%d] \r",count++,max); fflush(stdout); }
		if (!P[index(l-1,x,y,z)].is_zero_matrix())
		{
			if (!isin(l,x,y,z)) printf("ERROR: not in l %d x %d y %d z %d--> increase index() range",l,x,y,z);
			CP[index(l,x,y,z)] = M.mul(P[index(l-1,x,y,z)]);
			#ifdef use_BB
			for (int i=0;i<Q;i++) for (int j=0;j<Q;j++) CP[index(l,x,y,z)](i,j) = subs(CP[index(l,x,y,z)](i,j),list_BB);
			#endif
			simplify_matrix(CP[index(l,x,y,z)]);
		}
		#ifdef use_BB
		trP[index(l-1,x,y,z)] = subs(trP[index(l-1,x,y,z)],list_BB);
		#endif
		trP[index(l-1,x,y,z)] = expand(trP[index(l-1,x,y,z)]);
	}
	delete [] P;
}


void CORE::compute_EFDE_coefs()
{
	matrix locA = matrix(Q,Q);

	if (verbosity==1) { printf("compute_EFDE_coefs:\n"); fflush(stdout); }
	// dry run
	int max=0,count=0;
	for (int l=0;l<L;l++)
	for (int i3=-l;(i3<=l && DIM==3) || (i3<=-l);i3++)
	for (int i2=-l;(i2<=l && DIM>=2) || (i2<=-l);i2++)
	for (int i1=-l;i1<=l;i1++)
	if (isin(l,i1,i2,i3))
			max++;

	for (int l=0;l<L;l++)
	for (int i3=-l;(i3<=l && DIM==3) || (i3<=-l);i3++)
	for (int i2=-l;(i2<=l && DIM>=2) || (i2<=-l);i2++)
	for (int i1=-l;i1<=l;i1++)
	if (isin(l,i1,i2,i3))
	{
		if (verbosity==1) { printf("[%d/%d] \r",count++,max); fflush(stdout); }
		if (!CP[index(l,i1,i2,i3)].is_zero_matrix())
		{
			for (int q=0;q<Q;q++)
			{
				int k1=-CC[0][q];
				int k2=(DIM>=2) ? -CC[1][q] : 0;
				int k3=(DIM>=3) ? -CC[2][q] : 0;
				int i1k1=i1+k1;
				int i2k2=i2+k2;
				int i3k3=i3+k3;

				locA = (constvel) ? A : As[index(l,i1k1,i2k2,i3k3)];

				matrix out, aux;
				if (feqorder>0) // using symbolic meq
				{
					if (constvel)	aux = ( (e[q].mul_scalar(1-locA(q,q))).add( E[q].mul(locA).mul( const_feq )) );
					else		aux = ( (e[q].mul_scalar(1-locA(q,q))).add( E[q].mul(locA).mul( feq[index(l,i1k1,i2k2,i3k3)] )) );
				} else 
				{
					if (constvel) 	aux = ( (e[q].mul_scalar(1-locA(q,q))).add( E[q].mul(locA).mul(Mi).mul( const_meq )) );
					else		aux = ( (e[q].mul_scalar(1-locA(q,q))).add( E[q].mul(locA).mul(Mi).mul( meq[index(l,i1k1,i2k2,i3k3)] )) );
				}
				simplify_matrix(aux,Q,1);
				out = CP[index(l,i1,i2,i3)].mul(aux);
				for (int d=0;d<Q;d++) coefs_efde[d][0][index(l,i1k1,i2k2,i3k3)] += expand(out(d,0));
				
				// fill vecm
				if (!constvel)
				{
					aux = (constvel) ? 
						CP[index(l,i1,i2,i3)].mul(E[q]).mul(BMi)
						:
						CP[index(l,i1,i2,i3)].mul(E[q]).mul(dBMi[index(l,i1k1,i2k2,i3k3)]);
					for (int q1=0;q1<Q;q1++) 
					for (int q2=0;q2<Q;q2++)
						coefs_efde[q1][q2][index(l,i1k1,i2k2,i3k3)] += expand(aux(q1,q2));
				}
			}
		}
	}
	if (verbosity==1) printf("\n");


	max=0;
	for (int l=0;l<L;l++)
	for (int i3=-(l+1);(i3<=(l+1) && DIM==3) || (i3<=-(l+1));i3++)
	for (int i2=-(l+1);(i2<=(l+1) && DIM>=2) || (i2<=-(l+1));i2++)
	for (int i1=-(l+1);i1<=(l+1);i1++)
	if (isin(l,i1,i2,i3))
		max++;

	count=0;
	if (verbosity==1) { printf("compute_EFDE_coefs: expanding:\n"); fflush(stdout); }
	for (int l=0;l<L;l++)
	for (int i3=-(l+1);(i3<=(l+1) && DIM==3) || (i3<=-(l+1));i3++)
	for (int i2=-(l+1);(i2<=(l+1) && DIM>=2) || (i2<=-(l+1));i2++)
	for (int i1=-(l+1);i1<=(l+1);i1++)
	if (isin(l,i1,i2,i3))
	{
		if (verbosity==1) { printf("[%d/%d] \r",count++,max); fflush(stdout); }
		ex c = expand(trP[index(l,i1,i2,i3)]);
		for (int d=0;d<Q;d++) coefs_efde[d][d][index(l,i1,i2,i3)] = expand(coefs_efde[d][d][index(l,i1,i2,i3)] + c);
	}
	if (verbosity==1) printf("\n");
}

void CORE::compute_EPDE_coefs()
{
	// EPDE coefficients "gamma"
	for (int k=0;k<=ORDER;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	if (j1+j2+j3==j)
	{
		if (verbosity==1) { printf("EPDE coefs order %d of %d: %d-%d-%d-%d\n",k,ORDER,k-j,j1,j2,j3); fflush(stdout); }
		ex glkoef = (ex)1;
		if (k-j>0) glkoef /= factorial(k-j);
		if (j1>0) glkoef /= factorial(j1);
		if (j2>0) glkoef /= factorial(j2);
		if (j3>0) glkoef /= factorial(j3);

		for (int q1=0;q1<Q;q1++)
		for (int q2=0;q2<Q;q2++)
			 coefs_epde[q1][q2][k-j][indO(j1,j2,j3)]=0;

		for (int l=0;l<L;l++)
		for (int i3=-l-1;(i3<=l+1 && DIM==3) || (i3<=-l-1);i3++)
		for (int i2=-l-1;(i2<=l+1 && DIM>=2) || (i2<=-l-1);i2++)
		for (int i1=-l-1;i1<=l+1;i1++)
		if (isin(l,i1,i2,i3))
		{
			ex koef = (ex)1;
			if (l>0) koef *= pow(-(ex)l,k-j);
			if (l==0 && k-j>0) koef*=0;
			if (j1>0 && i1!=0) koef *= pow((ex)i1,j1);
			if (i1==0 && j1>0) koef *= 0;
			if (j2>0 && i2!=0) koef *= pow((ex)i2,j2);
			if (i2==0 && j2>0) koef *= 0;
			if (j3>0 && i3!=0) koef *= pow((ex)i3,j3);
			if (i3==0 && j3>0) koef *= 0;
			for (int q1=0;q1<Q;q1++) 
			for (int q2=0;q2<Q;q2++)
				coefs_epde[q1][q2][k-j][indO(j1,j2,j3)] += coefs_efde[q1][q2][index(l,i1,i2,i3)]*koef;
		}

		// compensate for LHS Taylor expansion of phi
		if (j==0) 
			for (int d=0;d<Q;d++) coefs_epde[d][d][k-j][indO(j1,j2,j3)] -= (ex)1;
		for (int q1=0;q1<Q;q1++)
		for (int q2=0;q2<Q;q2++)
			coefs_epde[q1][q2][k-j][indO(j1,j2,j3)] *= glkoef;
	}

	if (verbosity==1) { printf("compute_EPDE_coefs: mass simplify coefs\n"); fflush(stdout);}

	// expand all
	for (int k=0;k<=ORDER;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	if (j1+j2+j3==j)
	for (int q1=0;q1<Q;q1++)
	for (int q2=0;q2<Q;q2++)
		coefs_epde[q1][q2][k-j][indO(j1,j2,j3)] = expand(coefs_epde[q1][q2][k-j][indO(j1,j2,j3)]);
}



void CORE::export_EFDE_coefs_bigprint(ex &c, ofstream &gout, char* str, bool last)
{
	COEFS CF;
	CF.basis.push_back(lbm.sym_o);
	for (int q=0;q<Q;q++) CF.basis.push_back(lbm.sym_os[q]);
	const int morder=6;
	CF.fill(c);
	gout << "\n${\\color{blue}"<<str<<"} = ";

	char ss[100];
	bool first=true;
	for (int order=0;order<=morder;order++)
	{
		for (int i=0;i<CF.v.size();i++)
		if (CF.v[i].order()==order)
		{
			sprintf(ss,"");
			if (order==0)
				sprintf(ss,"1");
			else
			{
				for (int q=0;q<Q;q++)
				if (CF.v[i].exp[q]!=0)
				{
					if (CF.v[i].exp[q]==1) 
						sprintf(ss,"%s\\omega_{%d}",ss,q);
					else
						sprintf(ss,"%s\\omega_{%d}^{%d}",ss,q,CF.v[i].exp[q]);
				}
			}
			if (!first)  gout<<" + ";  else first=false;
			gout<<"{\\color{black} \\beta_{["<<ss<<"]} }";
			if (order>0) gout<<"{"<<ss<<"}";
		}
	}
	gout<<",$\n\nwhere\n\n";

	// now plot each coefficient
	bool firstone=true;
	int count2=0;
	int last2=0;
	for (int order=0;order<=morder;order++)
	{
		for (int i=0;i<CF.v.size();i++)
		if (CF.v[i].order()==order)
		{
			last2=count2;
			count2++;
		}
	}

	count2=0;
	for (int order=0;order<=morder;order++)
	{
		for (int i=0;i<CF.v.size();i++)
		if (CF.v[i].order()==order)
		{
			sprintf(ss,"");
			if (order==0)
				sprintf(ss,"1");
			else
			{
				for (int q=0;q<Q;q++)
				if (CF.v[i].exp[q]!=0)
				{
					if (CF.v[i].exp[q]==1) 
						sprintf(ss,"%s\\omega_{%d}",ss,q);
					else
						sprintf(ss,"%s\\omega_{%d}^{%d}",ss,q,CF.v[i].exp[q]);
				}
			}
			gout<<"$\\color{black}\\scriptscriptstyle ";
			gout<<"{ \\color{red} \\beta_{["<<ss<<"]} } = ";
			ostringstream s1;
			s1 << latex;
			if (factor_coefs)
				s1 << factor(CF.v[i].coef);
			else
				s1 << expand(CF.v[i].coef);
			string str1 = ReplaceString( s1.str(), "{(", "(");
			string str2 = ReplaceString( str1, ")}", ")");
			gout <<str2;
			if (last && count2==last2) gout << "."; else gout<<",";
			gout<<"$\n\\vskip0.1em ";
			count2++;
		}
	}
	gout<<endl;
}


void CORE::export_EFDE_coefs(int fde)
{
	bool print_all_coefficients=false;
	char momsym[100], momsymno[100];
	int last_L_nonzero=0;
	bool first=true;

	mkdir(target_dir,0777);
	char fn[100];
	sprintf(fn,"%s/d%dq%dm%d%s_efde%d.tex",target_dir,DIM,Q,collision_model,(constvel) ? "c" : "n", fde);

	if (verbosity==1) { printf("export_EFDE_coefs: create %s\n", fn); fflush(stdout); }
	ofstream gout(fn);
	gout << latex;
	gout<< "\\begin{flushleft}" << endl;
	gout<< "\\noindent"<<endl;
	gout<< "\t\\begin{framed}"<<endl;
	gout<< "\t\\begin{dmath*}"<<endl;

	char gamma_str[100];
	switch (DIM) {
		case 1: sprintf(gamma_str," \\gamma^{t}_{x}"); break;
		case 2: sprintf(gamma_str," \\gamma^{t}_{x,y}"); break;
		case 3: sprintf(gamma_str," \\gamma^{t}_{x,y,z}"); break;
	}
	lbm.moment_symbol(momsym, fde, Q);
	lbm.moment_symbol_number(momsymno, fde, Q);
	char strr[100];
	if (DIM==1) sprintf(strr,"x");
	if (DIM==2) sprintf(strr,"x,y");
	if (DIM==3) sprintf(strr,"x,y,z");
	if (lbm.moment_symbol_number(momsymno, fde, Q))
		gout << "\t"<<momsym<<"^{t+\\dt}_{"<<momsymno<<","<<strr<<"} =";
	else
		gout << "\t"<<momsym<<"^{t+\\dt}_{"<<strr<<"} =";

	if (print_all_coefficients)
	{
		for (int l=0;l<L;l++)
		for (int i3=-l-1;(i3<=l+1 && DIM==3) || (i3<=-l-1);i3++)
		for (int i2=-l-1;(i2<=l+1 && DIM>=2) || (i2<=-l-1);i2++)
		for (int i1=-l-1;i1<=l+1;i1++)
		for (int q=0;q<Q;q++)
		if (isin(l,i1,i2,i3))
		{
			ex c = coefs_efde[fde][q][index(l,i1,i2,i3)];
			if (c!=0)
			{
				char s_space_coeff[100], s_time_coeff[100], s_space_formula[100], s_time_formula[100];
				desc2str(s_space_coeff, s_time_coeff, s_space_formula, s_time_formula, l, i1, i2, i3);
				if (first) first=false; else gout<< "+";
				gout << "\t\\alpha^{"<<s_time_formula<<"}_{"<<s_space_formula<<"} \\phi^{"<<s_time_formula<<"}_{"<<s_space_formula<<"}\n";
				last_L_nonzero=l;
			}
		}
	} else
	{
		bool first=true;
		for (int q=0;q<Q;q++)
		{
			int Lh=-1;
			for (int l=0;l<L;l++)
			for (int i3=-l-1;(i3<=l+1 && DIM==3) || (i3<=-l-1);i3++)
			for (int i2=-l-1;(i2<=l+1 && DIM>=2) || (i2<=-l-1);i2++)
			for (int i1=-l-1;i1<=l+1;i1++)
			if (isin(l,i1,i2,i3))
				if (expand(coefs_efde[fde][q][index(l,i1,i2,i3)])!=0) Lh=l;

			lbm.moment_symbol(momsym, q, Q);
			lbm.moment_symbol_number(momsymno, q, Q);

			if (Lh>=0)
			{
				if (first) first=false; else gout<<" + ";
				if (Lh==0) 
				{
					if (DIM==1) gout << "\\sum_{i=-1}^{1}";
					if (DIM==2) gout << "\\sum_{i,j=-1}^{1}";
					if (DIM==3) gout << "\\sum_{i,j,k=-1}^{1}";
					gout << "\\alpha^{["<<momsym<<"_{"<<momsymno<<"}],t}_{x+i\\dl";
					if (DIM>=2) gout <<",y+j\\dl";
					if (DIM>=3) gout <<",z+k\\dl";
					gout << "}\n";
					gout << momsym <<"^{t}";
				} else
				{
					gout << "\\sum_{\\ell=0}^{"<<Lh<<"}";
					if (DIM==1) gout << "\\sum_{i=-\\ell-1}^{\\ell+1}";
					if (DIM==2) gout << "\\sum_{i,j=-\\ell-1}^{\\ell+1}";
					if (DIM==3) gout << "\\sum_{i,j,k=-\\ell-1}^{\\ell+1}";
					gout << "\\alpha^{["<<momsym<<"_{"<<momsymno<<"}],t-\\ell\\dt}_{x+i\\dl";
					if (DIM>=2) gout <<",y+j\\dl";
					if (DIM>=3) gout <<",z+k\\dl";
					gout << "}\n";
					gout << momsym <<"^{t-\\ell\\dt}";
				}
				if (q==0)
					gout<<"_{x+i\\dl";
				else
					gout<<"_{"<<momsymno<<",x+i\\dl";
				if (DIM>=2) gout <<",y+j\\dl";
				if (DIM>=3) gout <<",z+k\\dl";
				gout << "}\n";;
			}
		}
	}
	gout << "\t,\n\\end{dmath*}\n";
	gout << "\\end{framed}\n";
	gout << "\twhere the non-zero coefficients are given by:\n\n";

	gout << "\\hangindent=1cm\n";
	for (int l=0;l<L;l++)
	for (int i3=-l-1;(i3<=l+1 && DIM==3) || (i3<=-l-1);i3++)
	for (int i2=-l-1;(i2<=l+1 && DIM>=2) || (i2<=-l-1);i2++)
	for (int i1=-l-1;i1<=l+1;i1++)
	if (isin(l,i1,i2,i3))
	for (int q=0;q<Q;q++)
	{
		ex c =coefs_efde[fde][q][index(l,i1,i2,i3)];
		if (c!=0)
		{
			char s_space_coeff[100], s_time_coeff[100], s_space_formula[100], s_time_formula[100], str[1024];
			desc2str(s_space_coeff, s_time_coeff, s_space_formula, s_time_formula, l, i1, i2, i3);

			lbm.moment_symbol(momsym, q, Q);
			lbm.moment_symbol_number(momsymno, q, Q);

			sprintf(str,"\\alpha^{[%s_{%s}],%s}_{%s}",momsym,momsymno,s_time_formula,s_space_formula);

			bool last=(l==last_L_nonzero);

			gout<<"${\\color{blue}"<<str<<"} = ";
			ostringstream s;
			s << latex;
			if (factor_coefs)
				s << factor(c);
			else
				s << expand(c);
			string str1 = ReplaceString( s.str(), "{(", "(");
			string str2 = ReplaceString( str1, ")}", ")");
			gout <<str2;
			if (last) gout<<"."; else gout<<",";
			gout<<"$\n\n\\vskip0.5em\n";
		}
	}

	gout << "\\end{flushleft}\n\\hangindent=0cm\n";
	gout.close();
	cout << dflt;
}


void CORE::export_EPDE_coefs_bigprint(ex &expr, ofstream &fout, char* str)
{
	COEFS CF;
	fill_sym_vels_to_basis(CF);
	const int morder=10;
	CF.order=morder;
	CF.fill(expr);
	CF.check(expr);
	fout << "${\\color{blue}"<<str<<"} = ";
	char ss[100];
	bool first=true;
	for (int order=0;order<=morder;order++)
	{
		for (int i=0;i<CF.v.size();i++)
		if (CF.v[i].coef!=0)
		if (CF.v[i].order()==order)
		{
			if (!first)  fout << " + ";  else first=false;
			fout << "{\\color{black} \\beta_{\\left[";
			// fill in the basis with powers
			bool none=true;
			for (int b=0;b<CF.basis.size();b++)
			{
				if (CF.v[i].exp[b] != 0)
				{
					if (CF.v[i].exp[b]==1)
						fout << CF.basis_tex[b];
					else
						fout << CF.basis_tex[b]<<"^{"<<CF.v[i].exp[b]<<"}";
					none=false;
				}
			}
			if (none) fout << "1";
			fout << "\\right]} }";
			
						// put dl^pow dt^pow ... count all powers
			int pow_dl=0;
			int pow_dt=0;

			if (order>0) 
			{
				for (int b=0;b<CF.basis.size();b++)
				if (CF.v[i].exp[b] > 0)
				{
					pow_dl+=CF.pow_dl[b]*CF.v[i].exp[b];
					pow_dt+=CF.pow_dt[b]*CF.v[i].exp[b];
				}
				// put dl^pow dt^pow
				if (pow_dt==1) fout << "\\dt ";
				if (pow_dt>1)  fout << "\\dt^{"<< pow_dt <<"} ";
				if (pow_dl==1) fout << "\\dl ";
				if (pow_dl>1)  fout << "\\dl^{"<< pow_dl <<"} ";

				for (int b=0;b<CF.basis.size();b++)
				{
					if (CF.v[i].exp[b] != 0)
					{
						if (CF.v[i].exp[b]==1)
							fout << CF.basis[b];
						else
							fout << "\\left( "<<CF.basis[b]<<"\\right)^{"<<CF.v[i].exp[b]<<"}";
					}
				}
			}
		}
	}
	fout << ",$\n\nwhere\n\n";

	// now plot each coefficient
	for (int order=0;order<=morder;order++)
	{
		for (int i=0;i<CF.v.size();i++)
		if (CF.v[i].coef!=0)
		if (CF.v[i].order()==order)
		{
			fout<<"$\\color{black}\\scriptstyle ";
			fout<<"{ \\color{red} \\beta_{\\left[";
			// fill in the basis with powers
			bool none=true;
			for (int b=0;b<CF.basis.size();b++)
			{
				if (CF.v[i].exp[b] != 0)
				{
					if (CF.v[i].exp[b]==1)
						fout << CF.basis_tex[b];
					else
						fout << CF.basis_tex[b]<<"^{"<<CF.v[i].exp[b]<<"}";
					none=false;
				}
			}
			if (none) fout << "1";
			fout << "\\right]} } =";
			// export coef into a omega-basis
			COEFS CF2; 
			CF2.basis.push_back(lbm.sym_o);
			for (int q=0;q<Q;q++) CF2.basis.push_back(lbm.sym_os[q]);
			CF2.order=morder;
			CF2.fill(CF.v[i].coef);
			CF2.check(CF.v[i].coef);
			ex outex=0;

			for (int or2=0;or2<=morder;or2++)
			for (int i2=0;i2<CF2.v.size();i2++)
			if (CF2.v[i2].order()==or2)
				outex += CF2.v[i2].coef * CF2.v[i2].base;

			outex = expand(outex);

			// eliminate \dl and \dt from the coefficient
			outex = subs(outex, sym_dl==1);
			outex = subs(outex, sym_dt==1);

			ostringstream s1;
			s1 << latex;
			if (factor_coefs)
				s1 << factor(outex);
			else
				s1 << expand(outex);
			string str1 = ReplaceString( s1.str(), "{(", "(");
			string str2 = ReplaceString( str1, ")}", ")");
			fout <<str2;
			fout<<"$\n\\vskip0.1em ";
		}
	}
	fout << "\n";
}

void CORE::export_EPDE_coefs(int pde)
{
	// prepare subs lists
	lst list_meqs, list_vels;
	if (use_tay_meqs) taylor_list_meqs(list_meqs);
	if (use_tay_vels) taylor_list_vels(list_vels);

	mkdir(target_dir,0777);
	bool first=true;
	int last_c_nonzero=0;
	int count=0;
	char fn[100];
	sprintf(fn,"%s/d%dq%dm%d%s_epde%d.tex",target_dir,DIM,Q,collision_model,(constvel) ? "c" : "n", pde);

	if (verbosity==1) { printf("export_EPDE_coefs: create %s\n", fn); fflush(stdout); }
	ofstream gout(fn);
	gout << latex;
	gout<< "\\begin{flushleft}" << endl;
	gout<< "\\noindent"<<endl;

	 // OVERVIEW
	gout<< "\t\\begin{framed}"<<endl;
	gout<< "\t\\begin{dmath*}"<<endl;

	for (int k=0;k<=ORDER;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	for (int q=0;q<Q;q++)
	if (j1+j2+j3==j)
	{
		if (coefs_epde[pde][q][k-j][indO(j1,j2,j3)]!=0)
		{
			gout << "\t";
			if (!first) gout << "+"; else first=false;
			char str[100], str_dl[50], str_dt[50];
			desc2str(str,k,j1,j2,j3);
			char gamma_str[100], momsym[100], momsymno[100];
			lbm.moment_symbol(momsym,q,Q);
			lbm.moment_symbol_number(momsymno,q,Q);
			sprintf(gamma_str,"\\gamma^{[%s_{%s}]}_{[%s]}",momsym,momsymno,str);
			gout << gamma_str;
			sprintf(str,"");
			if (k-j>0) { if (k-j==1) sprintf(str_dt,"\\dt"); else sprintf(str_dt,"\\dt^{%d}",k-j); } else sprintf(str_dt,"");
			if (j>0) { if (j==1) sprintf(str_dl,"\\dl"); else sprintf(str_dl,"\\dl^{%d}",j); } else sprintf(str_dl,"");
			
			// put dl^pow dt^pow
			int pow_dl=tay_ms[q]->get_pow_dl(k-j,j1,j2,j3);
			int pow_dt=tay_ms[q]->get_pow_dt(k-j,j1,j2,j3);
			if (pow_dt==1) gout << "\\dt ";
			if (pow_dt>1) gout << "\\dt^{"<< pow_dt <<"} ";
			if (pow_dl==1) gout << "\\dl ";
			if (pow_dl>1) gout << "\\dl^{"<< pow_dl <<"} ";
			
			gout << tay_ms[q]->get_tay(k-j,j1,j2,j3) << endl;
			last_c_nonzero=count;
			count++;
		}
	}
	gout<< "\t=0,\\end{dmath*}"<<endl;
	gout<< "\\end{framed}\n\twhere\n\n";
	gout<<"\\hangindent=1cm\n";

	
	int max=0;

	// dry run
	for (int k=0;k<=ORDER;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	for (int q=0;q<Q;q++)
	if (j1+j2+j3==j)
		if (coefs_epde[pde][q][k-j][indO(j1,j2,j3)]!=0)
			max++;
	
	count=0;
	for (int k=0;k<=ORDER;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	for (int q=0;q<Q;q++)
	if (j1+j2+j3==j)
	{
		if (coefs_epde[pde][q][k-j][indO(j1,j2,j3)]!=0)
		{
			count++;
			if (verbosity==1) { printf("[%d/%d]  \r",count,max); fflush(stdout); }

			ex expr = expand(coefs_epde[pde][q][k-j][indO(j1,j2,j3)]);
			if (use_tay_meqs) expr = expand(subs(expr, list_meqs));
			if (use_tay_vels) expr = expand(subs(expr, list_vels));
			char str[100], gamma_str[100], momsym[100], momsymno[100];
			desc2str(str,k,j1,j2,j3);
			lbm.moment_symbol(momsym,q,Q);
			lbm.moment_symbol_number(momsymno,q,Q);
			sprintf(gamma_str,"\\gamma^{[%s_{%s}]}_{[%s]}",momsym,momsymno,str);
			if (bigprint)
			{
				export_EPDE_coefs_bigprint(expr, gout, gamma_str);
			} else 
			{
				gout << "${\\color{blue}"<<gamma_str<<"} = ";
				ostringstream s1;
				s1 << latex;
				if (factor_coefs)
					s1 << factor(expr);
				else
					s1 << expand(expr);
				string str1 = ReplaceString( s1.str(), "{(", "(");
				string str2 = ReplaceString( str1, ")}", ")");
				gout <<str2;
				gout <<  ((count==last_c_nonzero) ?  "." : ",");
				gout << "$\n\n\\vskip0.5em\n";
			}
		}
	}
	gout << "\\end{flushleft}\n\\hangindent=0cm\n";
	gout.close();
}


CORE::~CORE()
{
	if (Feq_allocated)
	{
		delete [] feq;
		delete [] meq;
	}
	if (allocated)
	{
		delete [] As;
		delete [] dBMi;

		delete [] CP;
		delete [] trP;

		if (use_tay_vels) { for (int d=0;d<DIM;d++) delete tay_vels[d]; delete [] tay_vels; }
		if (use_tay_ms) { for (int q=0;q<Q;q++) delete tay_ms[q]; delete [] tay_ms; }
		if (use_tay_meqs) { for (int q=0;q<Q;q++) delete tay_meqs[q]; delete [] tay_meqs; }

		for (int q=0;q<Q;q++)
		{
			delete [] sym_fs[q];
			delete [] sym_ms[q];
			delete [] sym_feqs[q];
			delete [] sym_meqs[q];
		}
		delete [] sym_fs;
		delete [] sym_ms;
		delete [] sym_feqs;
		delete [] sym_meqs;

		for (int d=0;d<DIM;d++) delete [] sym_vels[d];
		delete [] sym_vels;

		for (int d=0;d<Q;d++)
		{
			for (int q=0;q<Q;q++)
			{
				 delete [] coefs_efde[d][q];
				 for (int o=0;o<=ORDER;o++) delete [] coefs_epde[d][q][o];
				 delete [] coefs_epde[d][q];
			}
			delete [] coefs_efde[d];
			delete [] coefs_epde[d];
		}
		delete [] coefs_efde;
		delete [] coefs_epde;

		delete [] sym_feq;
		delete [] sym_f;

		delete [] sym_meq;
		delete [] sym_m;
	}
}
