//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"
#include "lbm.h"
#include "coefs.h"
#include "taylor.h"
#include "cs.h"

struct CORE : DEF
{
	LBM lbm;

	int ORDER=2;	// Taylor order
	int collision_model=0;
	int feqorder=-1; // negative = symbolic
	int verbosity=0;
	bool constvel=false;
	bool bigprint;
	bool EFDEprint;
	bool use_tay_vels=true;
	bool use_tay_meqs=true;
	bool use_tay_ms=true;
	bool use_symbolic_B=false;
	bool factor_coefs=true;

	char target_dir[1024];

	matrix I, E[MQ], A, B, C, M, Mi, BMi, e[MQ];
	matrix *As, *dBMi;
	vector<CS> vecCS;
	int CC[MDIM][MQ];
	matrix *CP;
	ex *trP;
	bool allocated=false;

	ex ***coefs_efde, ****coefs_epde;
	ex **sym_vels, **sym_fs, **sym_ms, **sym_feqs, **sym_meqs;
	// central values:
	ex *sym_m, *sym_meq, *sym_f, *sym_feq;

	ex sym_dl=symbol("\\dl");
	ex sym_dt=symbol("\\dt");

	TAYLOR **tay_vels, **tay_ms, **tay_meqs;

	matrix const_feq, const_sub_feq, const_meq;
	matrix *feq, *meq;
	bool Feq_allocated=false;

	ex sym_cs, sym_vel[MDIM];

	void taylor_list_vels(lst &list);
	void taylor_list_fs(lst &list);
	void taylor_list_ms(lst &list);
	void taylor_list_meqs(lst &list);


	void desc2str(char*str, int k, int j1, int j2, int j3);
	void number2str(char*str, const char*s1, const char*s2, int number);
	void coord2str(char*str, int x, int y, int z);
	void desc2str(char *s_space_coeff, char* s_time_coeff, char* s_space_formula, char* s_time_formula, int l, int i1, int i2, int i3);

	std::string ReplaceString(std::string subject, const std::string& search, const std::string& replace);
	inline void simplify_matrix(matrix &A) { simplify_matrix(A,Q,Q); }
	inline void simplify_matrix(matrix &A, int Nrows, int Ncols) { for (int i=0;i<Nrows;i++) for (int j=0;j<Ncols;j++) A(i,j) = expand(A(i,j)); }
	inline void empty_matrix(matrix &A) { empty_matrix(A,Q,Q); }
	inline void empty_matrix(matrix &A, int Nrows, int Ncols) { for (int i=0;i<Nrows;i++) for (int j=0;j<Ncols;j++) A(i,j)=0;; }

	inline int indO(int a, int b=0, int c=0) { return a + b*(ORDER+1) + c*(ORDER+1)*(ORDER+1); }
	void Meq_symbolic(matrix &M, int l, int x, int y, int z);
	void Feq_symbolic(matrix &F, int l, int x, int y, int z);
	void Feq_symbolic_Taylor(matrix &F, int l, int x, int y, int z);
	void fill_sym_vels_to_basis(COEFS &CF);

	void define_matrix_A_worker(matrix &_A, ex &u, ex &v, ex &w);
	void define_matrix_A();
	void define_matrix_B();
	void define_matrix_M();
	void define_equilibrium();
	void allocate();

	void compute_coefficients();
	void compute_EFDE_coefs();
	void compute_EPDE_coefs();

	void export_EFDE_coefs_bigprint(ex &c, ofstream &gout, char* str, bool last);
	void export_EFDE_coefs(int pde);
	void export_EPDE_coefs(int pde);
	void export_EPDE_coefs_bigprint(ex &expr, ofstream &fout, char *str);

	CORE(int iDIM, int iQ) : DEF(iDIM, iQ) { sprintf(target_dir,"default"); };
	~CORE();
};
