//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"

const int CS_MAX=MQ+1;

struct CS
{
	int k[CS_MAX];
	int N=0;
	bool overflow=false;

	ex e=0; // stored expression, possible computed beta

	bool increment(int i, int Max)
	{
		if (k[i]==Max)
		{
			if (i==N) { overflow=true; return false; } // overflow
			k[i]=0;
			return increment(i+1, Max);
		} else
		{
			k[i]++;
		}
		return true;
	}
	
	bool match(int *k_other, int n)
	{
		if (n!=N) return false;
		for (int i=0;i<=N;i++) if (k[i]!=k_other[i]) return false;
		return true;
	}

	void reset()
	{
		overflow=false;
		for (int i=0;i<=N;i++) k[i]=0; 
	}
	
	CS(int size) {
		N=CS_MAX-1;
		if (size>=CS_MAX-1) printf("Error: CS_MAX exceeded, increase CS_MAX to fit %d\n",size);
		else
			N=size;
		reset();
	}
};
