//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <sys/stat.h>
#include <ginac/ginac.h>

const int MQ=27;// maximal Q ... lazy memory allocation
const int MDIM=3; // maximal DIM .. lazy memory allocation

using namespace std;
using namespace GiNaC;

enum { 
	SRT,		// 0
	MRT1,		// 1
	MRT2,		// 2
	MRT1_OG,	// 3
	MRT2_OG,	// 4
	CLBM1,		// 5
	CLBM2,		// 6
	CLBM1_OG,	// 7
	CLBM2_OG,	// 8
	GENERAL		// 9
};

struct DEF
{
	bool internal_index=true;
	int R;
	int Q;
	int DIM;
	int L;

	int Nindex=0;  // total number of entries needed for the algorithm
	int **_index; // main index function (replaces ind)
	
	void allocate_index()
	{
		_index = new int*[L+1];
		for (int l=0;l<=L;l++)
		{
			_index[l] = new int[(2*R+1)*(2*R+1)*(2*R+1)];
			for (int i=0;i<(2*R+1)*(2*R+1)*(2*R+1);i++) _index[l][i]=-1;
		}
	}
	
	// test whether we are within the range
	bool isin(int l, int a, int b, int c)
	{
		return (_index[l][ind(a,b,c)]>=0);
	}
	
	int index(int l, int a, int b, int c)
	{
		if (_index[l][ind(a,b,c)] < 0)
		{ 
			printf("indexing error: out of range: %d of %d,  %d of %d,  %d of %d,  %d vs. %d\n", l,L,R+a,2*R,R+b,2*R,R+c,2*R);
			return 0;
		}
		return _index[l][ind(a,b,c)];
	}

	void fill_index()
	{
		int count=0;
		int ra=1; // plus one!
		for (int l=0;l<=L;l++)
		{
			if (l<=L/2)
				ra++;
			else 
				ra--;
			for (int z=-ra;(z<=ra && DIM==3) || (z<=-ra);z++)
			for (int y=-ra;(y<=ra && DIM>=2) || (y<=-ra);y++)
			for (int x=-ra;x<=ra;x++)
			{
				_index[l][ind(x,y,z)]=count;
				count++;
			}
		}
		Nindex=count; // max number of entries
	}


	int ind(int a, int b, int c) 
	{
		switch (DIM)
		{
			case 1: return (a+R);
			case 2: return (a+R)+(b+R)*(2*R+1);
			case 3: return (a+R)+(b+R)*(2*R+1)+(c+R)*(2*R+1)*(2*R+1);
		}
		return 0;
	}


	// constructor
	DEF(int iDIM, int iQ)
	{
		DIM=iDIM;
		Q=iQ;
		L=iQ+1;
		R=L+1;
		// allocated indexes
		allocate_index();
		fill_index();
		internal_index=true;
	}

	// constructor
	DEF(int iDIM, int iQ, int**external_index, int external_Nindex)
	{
		DIM=iDIM;
		Q=iQ;
		L=iQ+1;
		R=L+1;
		_index=external_index;
		Nindex = external_Nindex;
		internal_index=false;
	}	
	~DEF()
	{
		// delete index
		if (internal_index)
		{
			for (int l=0;l<=L;l++) delete [] _index[l];
			delete [] _index;
		}
	}
};
