//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"

struct LBM
{
	ex sym_o=symbol("\\omega");
	ex sym_os[MQ];

	const ex no1 = (ex)1;
	const ex n1o2 = 1/(ex)2;
	const ex n1o3 = 1/(ex)3;
	const ex n2o3 = 2/(ex)3;
	const ex n1o4 = 1/(ex)4;
	const ex n1o5 = 1/(ex)5;
	const ex n4o5 = 4/(ex)5;
	const ex n1o6 = 1/(ex)6;
	const ex n1o7 = 1/(ex)7;
	const ex n5o7 = 5/(ex)7;
	const ex n6o7 = 6/(ex)7;
	const ex n1o12 = 1/(ex)12;
	const ex n1o20 = 1/(ex)20;
	const ex n2o15 = 2/(ex)15;
	const ex n3o10 = 3/(ex)10;
	const ex n1o42 = 1/(ex)42;
	const ex n2o35 = 2/(ex)35;
	const ex n5o14 = 5/(ex)14;

	void setA_SRT(matrix &A, int Q, ex u, ex v, ex w);
	void setA_MRT_using_M(matrix &A, int Q, ex u, ex v, ex w);
	void setA_MRT1(matrix &A, int Q, ex u, ex v, ex w);
	void setA_MRT2(matrix &A, int Q, ex u, ex v, ex w);
	void setA_CLBM1(matrix &A, int Q, ex u, ex v, ex w);
	void setA_CLBM2(matrix &A, int Q, ex u, ex v, ex w);
	void setA_MRT1_OG(matrix &A, int Q, ex u, ex v, ex w);
	void setA_MRT2_OG(matrix &A, int Q, ex u, ex v, ex w);
	void setA_CLBM1_OG(matrix &A, int Q, ex u, ex v, ex w);
	void setA_CLBM2_OG(matrix &A, int Q, ex u, ex v, ex w);
	void setA_GENERAL(matrix &A, int Q, ex u, ex v, ex w);

	void setM_MRT1(matrix &M, int Q);
	void setM_MRT1_15(matrix &M);
	void setM_MRT1_19(matrix &M);
	void setM_MRT1_27(matrix &M);
	void setM_MRT2(matrix &M, int Q);
	
	void setC_3(matrix &C);
	void setC_5(matrix &C);
	void setC_9(matrix &C);
	void setC_7(matrix &C);
	void setC_15(matrix &C);
	void setC_19(matrix &C);
	void setC_27(matrix &C);
	void setC(matrix &C, int Q);
	
	void moment_symbol(char *desc, int q, int Q);
	bool moment_symbol_number(char *desc, int q, int Q);
	
	void Feq(matrix &F, ex u, ex v, ex w, ex &sym_cs, int Q, int order);
	
	LBM()
	{
		char desc[20];
		for (int q=0;q<MQ;q++)
		{
			sprintf(desc,"\\omega_{%d}",q);
			sym_os[q]=symbol(desc);
		}
	}

};
