//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"

struct TAYLOR : DEF
{
	int ORDER;
	ex *tay=0;
	ex *tay_tex=0;
	ex *sub=0;
	ex central;

	// this needs to be initialized
	ex dl;
	ex dt;
	
	// powers of \dl and \dt
	int *pow_dl=0;
	int *pow_dt=0;

	void allocate(const char *comp, ex central_value);
	void taylor_expand();

	int indO(int a, int b, int c, int d) { return a + b*(ORDER+1) + c*(ORDER+1)*(ORDER+1) + d*(ORDER+1)*(ORDER+1)*(ORDER+1); }

	ex& get_tay(int a, int b, int c, int d) { return tay[indO(a,b,c,d)]; }
	ex& get_tay_tex(int a, int b, int c, int d) { return tay_tex[indO(a,b,c,d)]; }

	int& get_pow_dl(int a, int b, int c, int d) { return pow_dl[indO(a,b,c,d)]; }
	int& get_pow_dt(int a, int b, int c, int d) { return pow_dt[indO(a,b,c,d)]; }

	TAYLOR(int iDIM, int iQ, int _ORDER, int **external_index, int external_Nindex, ex sym_dl, ex sym_dt) : DEF(iDIM, iQ, external_index, external_Nindex)
	{
		dl=sym_dl;
		dt=sym_dt;
		ORDER=_ORDER;
	}

	~TAYLOR()
	{
		if (tay) delete [] tay;
		if (tay_tex) delete [] tay_tex;
		if (sub) delete [] sub;
		if (pow_dl) delete [] pow_dl;
		if (pow_dt) delete [] pow_dt;
	}
};

