#include "lbm.h"

void LBM::setA_GENERAL(matrix &A, int Q, ex u, ex v, ex w)
{
	ex SA = symbolic_matrix(Q,Q,"a");
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
		A(i,j) = SA[j+i*Q]; // symbolic
}


void LBM::setA_SRT(matrix &A, int Q, ex u, ex v, ex w)
{
	// SRT: identity * omega 
	for (int i=0;i<Q;i++)
		A(i,i)=sym_o;
}

void LBM::setA_MRT_using_M(matrix &A, int Q,ex u,ex v,ex w)
{
	matrix M = matrix(Q,Q);
	setM_MRT1(M,Q);
	// prepare S matrix
	matrix S = matrix(Q,Q);
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
		S(i,j) = (i==j) ? sym_os[i] : 0;
	matrix Mi = inverse(M);
	A = Mi.mul(S).mul(M);
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
		A(i,j) = expand(A(i,j));
}


void LBM::setA_MRT1(matrix &A, int Q, ex u, ex v, ex w)
{
	switch (Q)
	{
	case 3:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[2];
		A(0,2) = sym_os[0]-sym_os[2];
		A(1,0) = 0;
		A(1,1) = n1o2*sym_os[2]+n1o2*sym_os[1];
		A(1,2) = n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,0) = 0;
		A(2,1) = n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,2) = n1o2*sym_os[2]+n1o2*sym_os[1];
		break;
	case 5:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[3];
		A(0,2) = sym_os[0]-sym_os[4];
		A(0,3) = sym_os[0]-sym_os[3];
		A(0,4) = sym_os[0]-sym_os[4];
		A(1,0) = 0;
		A(1,1) = n1o2*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = 0;
		A(1,3) = n1o2*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = 0;
		A(2,0) = 0;
		A(2,1) = 0;
		A(2,2) = n1o2*sym_os[4]+n1o2*sym_os[2];
		A(2,3) = 0;
		A(2,4) = n1o2*sym_os[4]-n1o2*sym_os[2];
		A(3,0) = 0;
		A(3,1) = n1o2*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = 0;
		A(3,3) = n1o2*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = 0;
		A(4,0) = 0;
		A(4,1) = 0;
		A(4,2) = n1o2*sym_os[4]-n1o2*sym_os[2];
		A(4,3) = 0;
		A(4,4) = n1o2*sym_os[4]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = sym_os[0];
		A(0,0) = sym_os[0]-sym_os[4];
		A(0,0) = sym_os[0]-sym_os[5];
		A(0,0) = sym_os[0]-sym_os[6];
		A(0,0) = sym_os[0]-sym_os[4];
		A(0,0) = sym_os[0]-sym_os[5];
		A(0,0) = sym_os[0]-sym_os[6];
		A(1,0) = 0;
		A(1,1) = n1o2*sym_os[4]+n1o2*sym_os[1];
		A(1,2) = 0;
		A(1,3) = 0;
		A(1,4) = n1o2*sym_os[4]-n1o2*sym_os[1];
		A(1,5) = 0;
		A(1,6) = 0;
		A(2,0) = 0;
		A(2,1) = 0;
		A(2,2) = n1o2*sym_os[5]+n1o2*sym_os[2];
		A(2,3) = 0;
		A(2,4) = 0;
		A(2,5) = n1o2*sym_os[5]-n1o2*sym_os[2];
		A(2,6) = 0;
		A(3,0) = 0;
		A(3,1) = 0;
		A(3,2) = 0;
		A(3,3) = n1o2*sym_os[6]+n1o2*sym_os[3];
		A(3,4) = 0;
		A(3,5) = 0;
		A(3,6) = n1o2*sym_os[6]-n1o2*sym_os[3];
		A(4,0) = 0;
		A(4,1) = n1o2*sym_os[4]-n1o2*sym_os[1];
		A(4,2) = 0;
		A(4,3) = 0;
		A(4,4) = n1o2*sym_os[4]+n1o2*sym_os[1];
		A(4,5) = 0;
		A(4,6) = 0;
		A(5,0) = 0;
		A(5,1) = 0;
		A(5,2) = n1o2*sym_os[5]-n1o2*sym_os[2];
		A(5,3) = 0;
		A(5,4) = 0;
		A(5,5) = n1o2*sym_os[5]+n1o2*sym_os[2];
		A(5,6) = 0;
		A(6,0) = 0;
		A(6,1) = 0;
		A(6,2) = 0;
		A(6,3) = n1o2*sym_os[6]-n1o2*sym_os[3];
		A(6,4) = 0;
		A(6,5) = 0;
		A(6,6) = n1o2*sym_os[6]+n1o2*sym_os[3];
		break;
	case 9:
		A(0,0)=sym_os[0];
		A(0,1)=sym_os[0]-sym_os[4];
		A(0,2)=sym_os[0]-sym_os[5];
		A(0,3)=sym_os[0]-sym_os[4];
		A(0,4)=sym_os[0]-sym_os[5];
		A(0,5)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(0,6)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(0,7)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(0,8)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(1,0)=0;
		A(1,1)=sym_os[4]/2+sym_os[1]/2;
		A(1,2)=0;
		A(1,3)=sym_os[4]/2-sym_os[1]/2;
		A(1,4)=0;
		A(1,5)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(1,6)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(1,7)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(1,8)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(2,0)=0;
		A(2,1)=0;
		A(2,2)=sym_os[5]/2+sym_os[2]/2;
		A(2,3)=0;
		A(2,4)=sym_os[5]/2-sym_os[2]/2;
		A(2,5)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(2,6)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(2,7)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(2,8)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(3,0)=0;
		A(3,1)=sym_os[4]/2-sym_os[1]/2;
		A(3,2)=0;
		A(3,3)=sym_os[4]/2+sym_os[1]/2;
		A(3,4)=0;
		A(3,5)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(3,6)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(3,7)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(3,8)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(4,0)=0;
		A(4,1)=0;
		A(4,2)=sym_os[5]/2-sym_os[2]/2;
		A(4,3)=0;
		A(4,4)=sym_os[5]/2+sym_os[2]/2;
		A(4,5)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(4,6)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(4,7)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(4,8)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(5,0)=0;
		A(5,1)=0;
		A(5,2)=0;
		A(5,3)=0;
		A(5,4)=0;
		A(5,5)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		A(5,6)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(5,7)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(5,8)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(6,0)=0;
		A(6,1)=0;
		A(6,2)=0;
		A(6,3)=0;
		A(6,4)=0;
		A(6,5)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(6,6)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		A(6,7)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(6,8)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(7,0)=0;
		A(7,1)=0;
		A(7,2)=0;
		A(7,3)=0;
		A(7,4)=0;
		A(7,5)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(7,6)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(7,7)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		A(7,8)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(8,0)=0;
		A(8,1)=0;
		A(8,2)=0;
		A(8,3)=0;
		A(8,4)=0;
		A(8,5)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(8,6)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(8,7)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(8,8)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		break;
	case 15:
	case 27: 
		setA_MRT_using_M(A,Q,u,v,w);
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and MRT1\n",Q);
	}
}

void LBM::setA_MRT2(matrix &A, int Q, ex u, ex v, ex w)
{
	switch (Q)
	{
	case 5:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[3];
		A(0,2) = sym_os[0]-sym_os[3];
		A(0,3) = sym_os[0]-sym_os[3];
		A(0,4) = sym_os[0]-sym_os[3];
		A(1,0) = 0;
		A(1,1) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(1,3) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(2,0) = 0;
		A(2,1) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(2,2) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		A(2,3) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(2,4) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(3,0) = 0;
		A(3,1) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(3,3) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(4,0) = 0;
		A(4,1) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(4,2) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(4,3) = n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,4) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[4];
		A(0,2) = sym_os[0]-sym_os[4];
		A(0,3) = sym_os[0]-sym_os[4];
		A(0,4) = sym_os[0]-sym_os[4];
		A(0,5) = sym_os[0]-sym_os[4];
		A(0,6) = sym_os[0]-sym_os[4];

		A(1,0) = 0;
		A(1,1) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6+sym_os[1]/2;
		A(1,2) = sym_os[4]/6-sym_os[5]/6;
		A(1,3) = sym_os[4]/6-sym_os[6]/6;
		A(1,4) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6-sym_os[1]/2;
		A(1,5) = sym_os[4]/6-sym_os[5]/6;
		A(1,6) = sym_os[4]/6-sym_os[6]/6;

		A(2,0) = 0;
		A(2,1) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(2,2) = sym_os[5]/3+sym_os[4]/6+sym_os[2]/2;
		A(2,3) = sym_os[4]/6-sym_os[6]/6;
		A(2,4) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(2,5) = sym_os[5]/3+sym_os[4]/6-sym_os[2]/2;
		A(2,6) = sym_os[4]/6-sym_os[6]/6;

		A(3,0) = 0;
		A(3,1) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(3,2) = sym_os[4]/6-sym_os[5]/6;
		A(3,3) = sym_os[6]/3+sym_os[4]/6+sym_os[3]/2;
		A(3,4) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(3,5) = sym_os[4]/6-sym_os[5]/6;
		A(3,6) = sym_os[6]/3+sym_os[4]/6-sym_os[3]/2;

		A(4,0) = 0;
		A(4,1) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6-sym_os[1]/2;
		A(4,2) = sym_os[4]/6-sym_os[5]/6;
		A(4,3) = sym_os[4]/6-sym_os[6]/6;
		A(4,4) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6+sym_os[1]/2;
		A(4,5) = sym_os[4]/6-sym_os[5]/6;
		A(4,6) = sym_os[4]/6-sym_os[6]/6;

		A(5,0) = 0;
		A(5,1) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(5,2) = sym_os[5]/3+sym_os[4]/6-sym_os[2]/2;
		A(5,3) = sym_os[4]/6-sym_os[6]/6;
		A(5,4) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(5,5) = sym_os[5]/3+sym_os[4]/6+sym_os[2]/2;
		A(5,6) = sym_os[4]/6-sym_os[6]/6;

		A(6,0) = 0;
		A(6,1) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(6,2) = sym_os[4]/6-sym_os[5]/6;
		A(6,3) = sym_os[6]/3+sym_os[4]/6-sym_os[3]/2;
		A(6,4) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(6,5) = sym_os[4]/6-sym_os[5]/6;
		A(6,6) = sym_os[6]/3+sym_os[4]/6+sym_os[3]/2;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and MRT2\n",Q);
	}
}

void LBM::setA_CLBM1(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	case 3:
		A(0,0) = (-sym_os[2]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[2]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[2]-2*sym_os[1])*u-sym_os[2]+sym_os[0];
		A(0,2) = (-sym_os[2]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[2])*u-sym_os[2]+sym_os[0];
		A(1,0) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(1,1) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[2]+sym_os[1]+sym_os[0])*u+n1o2*sym_os[2]+n1o2*sym_os[1];
		A(1,2) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[2]-3*sym_os[1]+sym_os[0])*u+n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,0) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(2,1) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[2]+3*sym_os[1]-sym_os[0])*u+n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,2) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[2]-sym_os[1]-sym_os[0])*u+n1o2*sym_os[2]+n1o2*sym_os[1];
		break;
	case 5:
		A(0,0) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[1])*u-sym_os[3]+sym_os[0];
		A(0,2) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[4]-2*sym_os[2])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,3) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[3])*u-sym_os[3]+sym_os[0];
		A(0,4) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[4])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(1,0) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(1,1) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[3]+sym_os[1]+sym_os[0])*u+n1o2*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(1,3) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[3]-3*sym_os[1]+sym_os[0])*u+n1o2*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(2,0) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[0]-sym_os[2])*v;
		A(2,1) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[0]-sym_os[2])*v;
		A(2,2) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(-2*sym_os[4]+sym_os[2]+sym_os[0])*v+n1o2*sym_os[4]+n1o2*sym_os[2];
		A(2,3) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[0]-sym_os[2])*v;
		A(2,4) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(2*sym_os[4]-3*sym_os[2]+sym_os[0])*v+n1o2*sym_os[4]-n1o2*sym_os[2];
		A(3,0) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(3,1) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[3]+3*sym_os[1]-sym_os[0])*u+n1o2*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(3,3) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[3]-sym_os[1]-sym_os[0])*u+n1o2*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(4,0) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[2]-sym_os[0])*v;
		A(4,1) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[2]-sym_os[0])*v;
		A(4,2) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(-2*sym_os[4]+3*sym_os[2]-sym_os[0])*v+n1o2*sym_os[4]-n1o2*sym_os[2];
		A(4,3) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[2]-sym_os[0])*v;
		A(4,4) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(2*sym_os[4]-sym_os[2]-sym_os[0])*v+n1o2*sym_os[4]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[4]-2*sym_os[1])*u-sym_os[4]+sym_os[0];
		A(0,2) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[5]-2*sym_os[2])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[5]+sym_os[0];
		A(0,3) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[6]-2*sym_os[3])*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[6]+sym_os[0];
		A(0,4) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[4])*u-sym_os[4]+sym_os[0];
		A(0,5) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[5])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[5]+sym_os[0];
		A(0,6) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[3]-2*sym_os[6])*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[6]+sym_os[0];

		A(1,0) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,1) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(-2*sym_os[4]+sym_os[1]+sym_os[0])*u+sym_os[4]+sym_os[1])*n1o2;
		A(1,2) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,3) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,4) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(2*sym_os[4]-3*sym_os[1]+sym_os[0])*u+sym_os[4]-sym_os[1])*n1o2;
		A(1,5) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,6) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;

		A(2,0) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,1) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,2) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(-2*sym_os[5]+sym_os[2]+sym_os[0])*v+sym_os[5]+sym_os[2])*n1o2;
		A(2,3) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,4) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,5) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(2*sym_os[5]-3*sym_os[2]+sym_os[0])*v+sym_os[5]-sym_os[2])*n1o2;
		A(2,6) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;

		A(3,0) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,1) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,2) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,3) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(-2*sym_os[6]+sym_os[3]+sym_os[0])*w+sym_os[6]+sym_os[3])*n1o2;
		A(3,4) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,5) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,6) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(2*sym_os[6]-3*sym_os[3]+sym_os[0])*w+sym_os[6]-sym_os[3])*n1o2;

		A(4,0) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,1) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(-2*sym_os[4]+3*sym_os[1]-sym_os[0])*u+sym_os[4]-sym_os[1])*n1o2;
		A(4,2) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,3) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,4) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(2*sym_os[4]-sym_os[1]-sym_os[0])*u+sym_os[4]+sym_os[1])*n1o2;
		A(4,5) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,6) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;

		A(5,0) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,1) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,2) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(-2*sym_os[5]+3*sym_os[2]-sym_os[0])*v+sym_os[5]-sym_os[2])*n1o2;
		A(5,3) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,4) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,5) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(2*sym_os[5]-sym_os[2]-sym_os[0])*v+sym_os[5]+sym_os[2])*n1o2;
		A(5,6) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;

		A(6,0) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,1) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,2) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,3) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(-2*sym_os[6]+3*sym_os[3]-sym_os[0])*w+sym_os[6]-sym_os[3])*n1o2;
		A(6,4) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,5) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,6) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(2*sym_os[6]-sym_os[3]-sym_os[0])*w+sym_os[6]+sym_os[3])*n1o2;
		break;
	case 9:
		A(0,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u-2*sym_os[0])/2;
		A(0,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[1]-4*sym_os[4])*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u-4*sym_os[4]+4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[4]-4*sym_os[1])*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+4*sym_os[4]-4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]-4*sym_os[4]+4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(0,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]-4*sym_os[4]+4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(0,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]+4*sym_os[4]-4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(0,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]+4*sym_os[4]-4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(1,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[0]-2*sym_os[1])*u)/4;
		A(1,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-sym_os[5])-3*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u+sym_os[5]+sym_os[4]+2*sym_os[1])/4;
		A(1,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[2])*u+2*sym_os[5]-2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(1,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+(4*sym_os[5]-2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(3*sym_os[5]+sym_os[4]-6*sym_os[1]+2*sym_os[0])*u+sym_os[5]+sym_os[4]-2*sym_os[1])/4;
		A(1,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[2])*u-2*sym_os[5]+2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(1,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-2*sym_os[5]-4*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(1,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+2*sym_os[5]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(1,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+2*sym_os[5]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(1,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-2*sym_os[5]-4*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(2,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u)/4;
		A(2,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+(2*sym_os[5]-2*sym_os[4])*u-sym_os[5]+sym_os[4])/4;
		A(2,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-sym_os[5]-3*sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+sym_os[4]+2*sym_os[2])/4;
		A(2,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+(2*sym_os[4]-2*sym_os[5])*u-sym_os[5]+sym_os[4])/4;
		A(2,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+4*sym_os[5]-2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+3*sym_os[5]+sym_os[4]-6*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]+sym_os[4]-2*sym_os[2])/4;
		A(2,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]-2*sym_os[5]-4*sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(2,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]-2*sym_os[5]-4*sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(2,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]+2*sym_os[5]-6*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(2,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]+2*sym_os[5]-6*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(3,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[0])*u)/4;
		A(3,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+(4*sym_os[5]-2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-3*sym_os[5])-sym_os[4]+6*sym_os[1]-2*sym_os[0])*u+sym_os[5]+sym_os[4]-2*sym_os[1])/4;
		A(3,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[2])*u+2*sym_os[5]-2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(3,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(sym_os[5]+3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u+sym_os[5]+sym_os[4]+2*sym_os[1])/4;
		A(3,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[2])*u-2*sym_os[5]+2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(3,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-2*sym_os[5]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(3,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+2*sym_os[5]+4*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(3,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+2*sym_os[5]+4*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(3,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-2*sym_os[5]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(4,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u)/4;
		A(4,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+(2*sym_os[5]-2*sym_os[4])*u-sym_os[5]+sym_os[4])/4;
		A(4,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+4*sym_os[5]-2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u-3*sym_os[5]-sym_os[4]+6*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]+sym_os[4]-2*sym_os[2])/4;
		A(4,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+(2*sym_os[4]-2*sym_os[5])*u-sym_os[5]+sym_os[4])/4;
		A(4,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+3*sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+sym_os[4]+2*sym_os[2])/4;
		A(4,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]-2*sym_os[5]+6*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(4,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]-2*sym_os[5]+6*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(4,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]+2*sym_os[5]+4*sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(4,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]+2*sym_os[5]+4*sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(5,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u)/8;
		A(5,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(5,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(5,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(5,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(5,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]+4*sym_os[5]+4*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(5,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(5,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]-4*sym_os[5]-4*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(5,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(6,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u)/8;
		A(6,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(6,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(6,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(6,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(6,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(6,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]-4*sym_os[5]-4*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(6,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(6,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]+4*sym_os[5]+4*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(7,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u)/8;
		A(7,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(7,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(7,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(7,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(7,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]-4*sym_os[5]-4*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(7,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(7,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]+4*sym_os[5]+4*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(7,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(8,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u)/8;
		A(8,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(8,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(8,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(8,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(8,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(8,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]+4*sym_os[5]+4*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(8,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(8,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]-4*sym_os[5]-4*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM1\n",Q);
	}
}

void LBM::setA_CLBM2(matrix &A, int Q, ex u, ex v, ex w)
{
	switch (Q)
	{
	case 5:
		A(0,0) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[1])*u-sym_os[3]+sym_os[0];
		A(0,2) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[3]-2*sym_os[2])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[3]+sym_os[0];
		A(0,3) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[3])*u-sym_os[3]+sym_os[0];
		A(0,4) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[3])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[3]+sym_os[0];
		A(1,0) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[1]-2*sym_os[0])*u;
		A(1,1) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[4]+2*sym_os[3]-2*sym_os[1]-2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[3]-2*sym_os[4])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[1]-2*sym_os[0])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(1,3) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(-2*sym_os[4]-2*sym_os[3]+6*sym_os[1]-2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[4]-2*sym_os[3])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[1]-2*sym_os[0])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(2,0) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[0]-2*sym_os[2])*v+n1o4*(sym_os[3]-sym_os[4])*u*u;
		A(2,1) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[0]-2*sym_os[2])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[4]-2*sym_os[3])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(2,2) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(-2*sym_os[4]-2*sym_os[3]+2*sym_os[2]+2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		A(2,3) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[0]-2*sym_os[2])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[3]-2*sym_os[4])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(2,4) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[4]+2*sym_os[3]-6*sym_os[2]+2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(3,0) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[0]-2*sym_os[1])*u;
		A(3,1) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[4]+2*sym_os[3]-6*sym_os[1]+2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[3]-2*sym_os[4])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[0]-2*sym_os[1])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(3,3) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(-2*sym_os[4]-2*sym_os[3]+2*sym_os[1]+2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[4]-2*sym_os[3])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[0]-2*sym_os[1])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,0) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u;
		A(4,1) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[4]-2*sym_os[3])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,2) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(-2*sym_os[4]-2*sym_os[3]+6*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(4,3) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[3]-2*sym_os[4])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,4) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[4]+2*sym_os[3]-2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[4]-2*sym_os[1])*u-sym_os[4]+sym_os[0];
		A(0,2) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[4]-2*sym_os[2])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,3) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[4]-2*sym_os[3])*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,4) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[4])*u-sym_os[4]+sym_os[0];
		A(0,5) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[4])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,6) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[3]-2*sym_os[4])*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];

		A(1,0) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u)/6;
		A(1,1) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(2*sym_os[6]+2*sym_os[5]+2*sym_os[4]-3*sym_os[1]-3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]-3*sym_os[1])/6;
		A(1,2) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[5]-sym_os[4])/6;
		A(1,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[6]-sym_os[4])/6;
		A(1,4) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(-2*sym_os[6]-2*sym_os[5]-2*sym_os[4]+9*sym_os[1]-3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]+3*sym_os[1])/6;
		A(1,5) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[5]-sym_os[4])/6;
		A(1,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[6]-sym_os[4])/6;

		A(2,0) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u)/6;
		A(2,1) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(2*sym_os[6]-4*sym_os[5]+2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(2,2) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(4*sym_os[5]+2*sym_os[4]-3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]-3*sym_os[2])/6;
		A(2,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;
		A(2,4) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(-2*sym_os[6]+4*sym_os[5]-2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(2,5) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(-4*sym_os[5]-2*sym_os[4]+9*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]+3*sym_os[2])/6;
		A(2,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;

		A(3,0) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u)/6;
		A(3,1) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(3,2) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(3,3) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(-4*sym_os[6]-2*sym_os[4]+3*sym_os[3]+3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]+3*sym_os[3])/6;
		A(3,4) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(-4*sym_os[6]+2*sym_os[5]+2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(3,5) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(3,6) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(4*sym_os[6]+2*sym_os[4]-9*sym_os[3]+3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]-3*sym_os[3])/6;

		A(4,0) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u)/6;
		A(4,1) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(2*sym_os[6]+2*sym_os[5]+2*sym_os[4]-9*sym_os[1]+3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]+3*sym_os[1])/6;
		A(4,2) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[5]-sym_os[4])/6;
		A(4,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[6]-sym_os[4])/6;
		A(4,4) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(-2*sym_os[6]-2*sym_os[5]-2*sym_os[4]+3*sym_os[1]+3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]-3*sym_os[1])/6;
		A(4,5) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[5]-sym_os[4])/6;
		A(4,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[6]-sym_os[4])/6;

		A(5,0) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u)/6;
		A(5,1) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(2*sym_os[6]-4*sym_os[5]+2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(5,2) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(4*sym_os[5]+2*sym_os[4]-9*sym_os[2]+3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]+3*sym_os[2])/6;
		A(5,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;
		A(5,4) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(-2*sym_os[6]+4*sym_os[5]-2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(5,5) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(-4*sym_os[5]-2*sym_os[4]+3*sym_os[2]+3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]-3*sym_os[2])/6;
		A(5,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;

		A(6,0) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u)/6;
		A(6,1) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(6,2) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(6,3) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(-4*sym_os[6]-2*sym_os[4]+9*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]-3*sym_os[3])/6;
		A(6,4) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(-4*sym_os[6]+2*sym_os[5]+2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(6,5) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(6,6) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(4*sym_os[6]+2*sym_os[4]-3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]+3*sym_os[3])/6;
		break;
	case 9:
		A(0,0)=(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[2]-2*sym_os[0])*v*v+sym_os[4]*u*u*u*u+((-2*sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[0])/2;
		A(0,1)=(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[4]-4*sym_os[1])*u-2*sym_os[4]+2*sym_os[0])/2;
		A(0,2)=(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u-4*sym_os[2])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[0])/2;
		A(0,3)=(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[1]-4*sym_os[4])*u-2*sym_os[4]+2*sym_os[0])/2;
		A(0,4)=(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+4*sym_os[2])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[0])/2;
		A(0,5)=(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]-4*sym_os[2])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(0,6)=(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]-4*sym_os[2])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(0,7)=(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]+4*sym_os[2])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(0,8)=(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]+4*sym_os[2])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(1,0)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[0])*u)/4;
		A(1,1)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+((-2*sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u-sym_os[4]-2*sym_os[1])/4;
		A(1,2)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(4*sym_os[7]+2*sym_os[5]-4*sym_os[3]+4*sym_os[2])*u+2*sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])/4;
		A(1,3)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-sym_os[4])+6*sym_os[1]-2*sym_os[0])*u-sym_os[4]+2*sym_os[1])/4;
		A(1,4)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-4*sym_os[7])-2*sym_os[5]+4*sym_os[3]-4*sym_os[2])*u-2*sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])/4;
		A(1,5)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+4*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]-sym_os[5]+3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(1,6)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]+2*sym_os[5]-12*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-sym_os[5]-sym_os[4]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(1,7)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]-2*sym_os[5]+12*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-sym_os[5]-sym_os[4]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(1,8)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-4*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]-sym_os[5]+3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(2,0)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u)/4;
		A(2,1)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u-2*sym_os[6]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+2*sym_os[4]*u-sym_os[4])/4;
		A(2,2)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u-sym_os[5]-2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]-2*sym_os[2])/4;
		A(2,3)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u-2*sym_os[6]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u-2*sym_os[4]*u-sym_os[4])/4;
		A(2,4)=-(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+3*sym_os[5]+6*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+2*sym_os[2])/4;
		A(2,5)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]+2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(2,6)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(2,7)=-(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]+3*sym_os[5]+sym_os[4]+6*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(2,8)=-(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]+3*sym_os[5]+sym_os[4]+6*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(3,0)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[0]-2*sym_os[1])*u)/4;
		A(3,1)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(sym_os[4]-6*sym_os[1]+2*sym_os[0])*u-sym_os[4]+2*sym_os[1])/4;
		A(3,2)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-4*sym_os[7])-2*sym_os[5]+4*sym_os[3]-4*sym_os[2])*u+2*sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])/4;
		A(3,3)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+((-2*sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+((-3*sym_os[4])+2*sym_os[1]+2*sym_os[0])*u-sym_os[4]-2*sym_os[1])/4;
		A(3,4)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(4*sym_os[7]+2*sym_os[5]-4*sym_os[3]+4*sym_os[2])*u-2*sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])/4;
		A(3,5)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]-2*sym_os[5]+12*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+sym_os[5]+sym_os[4]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(3,6)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-4*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]+sym_os[5]-3*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(3,7)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+4*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]+sym_os[5]-3*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(3,8)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]+2*sym_os[5]-12*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+sym_os[5]+sym_os[4]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(4,0)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u)/4;
		A(4,1)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u+2*sym_os[6]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+2*sym_os[4]*u-sym_os[4])/4;
		A(4,2)=-(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-3*sym_os[5]-6*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+2*sym_os[2])/4;
		A(4,3)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u+2*sym_os[6]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u-2*sym_os[4]*u-sym_os[4])/4;
		A(4,4)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]-2*sym_os[2])/4;
		A(4,5)=-(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]-3*sym_os[5]-sym_os[4]-6*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(4,6)=-(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]-3*sym_os[5]-sym_os[4]-6*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(4,7)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(4,8)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]+2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(5,0)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u)/8;
		A(5,1)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u-sym_os[4]*u*u+sym_os[4]*u)/8;
		A(5,2)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(5,3)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+3*sym_os[4]*u*u+sym_os[4]*u)/8;
		A(5,4)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(5,5)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(5,6)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(5,7)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]-2*sym_os[5]+2*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(5,8)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(6,0)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u)/8;
		A(6,1)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+3*sym_os[4]*u*u-sym_os[4]*u)/8;
		A(6,2)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(6,3)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u-sym_os[4]*u*u-sym_os[4]*u)/8;
		A(6,4)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(6,5)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(6,6)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(6,7)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(6,8)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]+2*sym_os[5]-2*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(7,0)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u)/8;
		A(7,1)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+3*sym_os[4]*u*u-sym_os[4]*u)/8;
		A(7,2)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(7,3)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u-sym_os[4]*u*u-sym_os[4]*u)/8;
		A(7,4)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(7,5)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]-2*sym_os[5]+2*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(7,6)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(7,7)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(7,8)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(8,0)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u)/8;
		A(8,1)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u-sym_os[4]*u*u+sym_os[4]*u)/8;
		A(8,2)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(8,3)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+3*sym_os[4]*u*u+sym_os[4]*u)/8;
		A(8,4)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(8,5)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(8,6)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]+2*sym_os[5]-2*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(8,7)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(8,8)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM2\n",Q);
	}
}


void LBM::setA_MRT1_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	case 5:
		A(0,0) = n2o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(0,1) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(0,2) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(0,3) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(0,4) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(1,0) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(1,1) = n3o10*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,2) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(1,3) = n3o10*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,4) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(2,0) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(2,1) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(2,2) = n1o6*sym_os[4]+n2o15*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		A(2,3) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(2,4) = n1o6*sym_os[4]+n2o15*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(3,0) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(3,1) = n3o10*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,2) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(3,3) = n3o10*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,4) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(4,0) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(4,1) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(4,2) = n1o6*sym_os[4]+n2o15*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(4,3) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(4,4) = n1o6*sym_os[4]+n2o15*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		break;
	case 7:
		A(0,0) = (2*sym_os[6])/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,1) = sym_os[0]/7-sym_os[4]/7;
		A(0,2) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,3) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,4) = sym_os[0]/7-sym_os[4]/7;
		A(0,5) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,6) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;

		A(1,0) = sym_os[0]/7-sym_os[4]/7;
		A(1,1) = (5*sym_os[4])/14+sym_os[1]/2+sym_os[0]/7;
		A(1,2) = sym_os[0]/7-sym_os[4]/7;
		A(1,3) = sym_os[0]/7-sym_os[4]/7;
		A(1,4) = (5*sym_os[4])/14-sym_os[1]/2+sym_os[0]/7;
		A(1,5) = sym_os[0]/7-sym_os[4]/7;
		A(1,6) = sym_os[0]/7-sym_os[4]/7;

		A(2,0) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(2,1) = sym_os[0]/7-sym_os[4]/7;
		A(2,2) = (3*sym_os[5])/10+(2*sym_os[4])/35+sym_os[2]/2+sym_os[0]/7;
		A(2,3) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(2,4) = sym_os[0]/7-sym_os[4]/7;
		A(2,5) = (3*sym_os[5])/10+(2*sym_os[4])/35-sym_os[2]/2+sym_os[0]/7;
		A(2,6) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;

		A(3,0) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(3,1) = sym_os[0]/7-sym_os[4]/7;
		A(3,2) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(3,3) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[3]/2+sym_os[0]/7;
		A(3,4) = sym_os[0]/7-sym_os[4]/7;
		A(3,5) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(3,6) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35-sym_os[3]/2+sym_os[0]/7;

		A(4,0) = sym_os[0]/7-sym_os[4]/7;
		A(4,1) = (5*sym_os[4])/14-sym_os[1]/2+sym_os[0]/7;
		A(4,2) = sym_os[0]/7-sym_os[4]/7;
		A(4,3) = sym_os[0]/7-sym_os[4]/7;
		A(4,4) = (5*sym_os[4])/14+sym_os[1]/2+sym_os[0]/7;
		A(4,5) = sym_os[0]/7-sym_os[4]/7;
		A(4,6) = sym_os[0]/7-sym_os[4]/7;

		A(5,0) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(5,1) = sym_os[0]/7-sym_os[4]/7;
		A(5,2) = (3*sym_os[5])/10+(2*sym_os[4])/35-sym_os[2]/2+sym_os[0]/7;
		A(5,3) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(5,4) = sym_os[0]/7-sym_os[4]/7;
		A(5,5) = (3*sym_os[5])/10+(2*sym_os[4])/35+sym_os[2]/2+sym_os[0]/7;
		A(5,6) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;

		A(6,0) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(6,1) = sym_os[0]/7-sym_os[4]/7;
		A(6,2) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(6,3) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35-sym_os[3]/2+sym_os[0]/7;
		A(6,4) = sym_os[0]/7-sym_os[4]/7;
		A(6,5) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(6,6) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[3]/2+sym_os[0]/7;
		break;
	case 9:
		A(0,0)=(4*sym_os[8])/9+(2*sym_os[5])/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(0,1)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(0,2)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(0,3)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(0,4)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(0,5)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(0,6)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(0,7)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(0,8)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(1,0)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(1,1)=sym_os[8]/9+sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(1,2)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(1,3)=sym_os[8]/9-sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(1,4)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(1,5)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(1,6)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(1,7)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(1,8)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(2,0)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(2,1)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(2,2)=sym_os[8]/9+sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9+sym_os[2]/6+sym_os[0]/9;
		A(2,3)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(2,4)=sym_os[8]/9-sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9-sym_os[2]/6+sym_os[0]/9;
		A(2,5)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(2,6)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(2,7)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(2,8)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(3,0)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(3,1)=sym_os[8]/9-sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(3,2)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(3,3)=sym_os[8]/9+sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(3,4)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(3,5)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(3,6)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(3,7)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(3,8)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(4,0)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(4,1)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(4,2)=sym_os[8]/9-sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9-sym_os[2]/6+sym_os[0]/9;
		A(4,3)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(4,4)=sym_os[8]/9+sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9+sym_os[2]/6+sym_os[0]/9;
		A(4,5)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(4,6)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(4,7)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(4,8)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(5,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(5,1)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(5,2)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(5,3)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(5,4)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(5,5)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(5,6)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(5,7)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(5,8)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(6,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(6,1)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(6,2)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(6,3)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(6,4)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(6,5)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(6,6)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(6,7)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(6,8)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(7,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(7,1)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(7,2)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(7,3)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(7,4)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(7,5)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(7,6)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(7,7)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(7,8)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(8,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(8,1)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(8,2)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(8,3)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(8,4)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(8,5)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(8,6)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(8,7)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(8,8)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		break;
	default:
		printf("Error: matrix A not implemented forQ%d and MRT1_OG\n",Q);
	}
}

void LBM::setA_MRT2_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	case 5:
		A(0,0) = n4o5*sym_os[3]+n1o5*sym_os[0];
		A(0,1) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(0,2) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(0,3) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(0,4) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(1,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(1,1) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,2) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(1,3) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,4) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(2,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(2,1) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(2,2) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		A(2,3) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(2,4) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(3,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(3,1) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,2) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(3,3) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,4) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(4,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(4,1) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(4,2) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(4,3) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(4,4) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		break;
	case 7:
		A(0,0) = (6*sym_os[4]+sym_os[0])/7;
		A(0,1) = -(sym_os[4]-sym_os[0])/7;
		A(0,2) =-(sym_os[4]-sym_os[0])/7;
		A(0,3) =-(sym_os[4]-sym_os[0])/7;
		A(0,4) =-(sym_os[4]-sym_os[0])/7;
		A(0,5) =-(sym_os[4]-sym_os[0])/7;
		A(0,6) =-(sym_os[4]-sym_os[0])/7;

		A(1,0) =-(sym_os[4]-sym_os[0])/7;
		A(1,1) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[1]+12*sym_os[0])/84;
		A(1,2) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(1,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(1,4) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[1]+12*sym_os[0])/84;
		A(1,5) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(1,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(2,0) =-(sym_os[4]-sym_os[0])/7;
		A(2,1) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(2,2) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[2]+12*sym_os[0])/84;
		A(2,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(2,4) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(2,5) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[2]+12*sym_os[0])/84;
		A(2,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(3,0) =-(sym_os[4]-sym_os[0])/7;
		A(3,1) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,2) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,3) =(14*sym_os[6]+sym_os[4]+21*sym_os[3]+6*sym_os[0])/42;
		A(3,4) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,5) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,6) =(14*sym_os[6]+sym_os[4]-21*sym_os[3]+6*sym_os[0])/42;

		A(4,0) =-(sym_os[4]-sym_os[0])/7;
		A(4,1) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[1]+12*sym_os[0])/84;
		A(4,2) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(4,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(4,4) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[1]+12*sym_os[0])/84;
		A(4,5) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(4,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(5,0) =-(sym_os[4]-sym_os[0])/7;
		A(5,1) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(5,2) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[2]+12*sym_os[0])/84;
		A(5,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(5,4) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(5,5) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[2]+12*sym_os[0])/84;
		A(5,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(6,0) =-(sym_os[4]-sym_os[0])/7;
		A(6,1) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,2) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,3) =(14*sym_os[6]+sym_os[4]-21*sym_os[3]+6*sym_os[0])/42;
		A(6,4) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,5) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,6) =(14*sym_os[6]+sym_os[4]+21*sym_os[3]+6*sym_os[0])/42;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and MRT2_OG\n",Q);
	}
}

void LBM::setA_CLBM1_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM1_OG\n",Q);
	}
}

void LBM::setA_CLBM2_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM2_OG\n",Q);
	}
}



void LBM::setM_MRT1_15(matrix &A)
{
	int maxima_matrix[15][15] {
		{1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{-2,	-1,	-1,	-1,	-1,	-1,	-1,	1,	1,	1,	1,	1,	1,	1,	1},
		{16,	-4,	-4,	-4,	-4,	-4,	-4,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	1,	0,	0,	-1,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	-4,	0,	0,	4,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	-4,	0,	0,	4,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	-4,	0,	0,	4,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	2,	-1,	-1,	2,	-1,	-1,	0,	0,	0,	0,	0,	0,	0,	0},
		{0,	0,	1,	-1,	0,	1,	-1,	0,	0,	0,	0,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	-1,	1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	-1,	1,	1,	-1}
		};
	for (int i=0;i<15;i++)
	for (int j=0;j<15;j++)
		A(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setM_MRT1_19(matrix &A)
{
	int maxima_matrix[19][19] {
		{1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{-30,	-11,	-11,	-11,	-11,	-11,	-11,	8,	8,	8,	8,	8,	8,	8,	8,	8,	8,	8,	8},
		{12,	-4,	-4,	-4,	-4,	-4,	-4,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1},
		{0,	-4,	0,	0,	4,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	-4,	0,	0,	4,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0},
		{0,	0,	0,	-4,	0,	0,	4,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1},
		{0,	2,	-1,	-1,	2,	-1,	-1,	-2,	-2,	-2,	-2,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	1,	-1,	0,	1,	-1,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1},
		{0,	-4,	2,	2,	-4,	2,	2,	-2,	-2,	-2,	-2,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	-2,	2,	0,	-2,	2,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	1,	-1,	1,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	-1,	1,	-1,	1,	0,	0,	0,	0,	1,	-1,	1,	-1}
		};
	for (int i=0;i<19;i++)
	for (int j=0;j<19;j++)
		A(i,j) = (ex)(maxima_matrix[i][j]);
}


void LBM::setM_MRT1_27(matrix &A)
{
	int maxima_matrix[27][27] {
		{1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	1,	-1,	-1,	1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	-1,	1,	1,	-1},
		{0,	1,	0,	0,	1,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	1,	0,	0,	1,	0,	1,	1,	1,	1,	1,	1,	1,	1,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	1,	0,	0,	1,	1,	1,	1,	1,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	-1,	1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1}
		};
	for (int i=0;i<27;i++)
	for (int j=0;j<27;j++)
		A(i,j) = (ex)(maxima_matrix[i][j]);
}


void LBM::setM_MRT1(matrix &A, int Q)
{
	switch (Q)
	{
	case 3:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = -no1;
		A(2,0) = 0;	A(2,1) = no1;	A(2,2) = no1;
		break;
	case 5:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) =-no1;	A(1,4) = 0;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) =-no1;
		A(3,0) = 0;	A(3,1) = no1;	A(3,2) = 0;	A(3,3) = no1;	A(3,4) = 0;
		A(4,0) = 0;	A(4,1) = 0;	A(4,2) = no1;	A(4,3) = 0;	A(4,4) = no1;
		break;
	case 7:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;	A(0,5) = no1;	A(0,6) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) = 0;	A(1,4) =-no1;	A(1,5) = 0;	A(1,6) = 0;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) = 0;	A(2,5) =-no1;	A(2,6) = 0;
		A(3,0) = 0;	A(3,1) = 0;	A(3,2) = 0;	A(3,3) = no1;	A(3,4) = 0;	A(3,5) = 0;	A(3,6) = -no1;
		A(4,0) = 0;	A(4,1) = no1;	A(4,2) = 0;	A(4,3) = 0;	A(4,4) = no1;	A(4,5) = 0;	A(4,6) = 0;
		A(5,0) = 0;	A(5,1) = 0;	A(5,2) = no1;	A(5,3) = 0;	A(5,4) = 0;	A(5,5) = no1;	A(5,6) = 0;
		A(6,0) = 0;	A(6,1) = 0;	A(6,2) = 0;	A(6,3) = no1;	A(6,4) = 0;	A(6,5) = 0;	A(6,6) = no1;
		break;
	case 9:
//	matrix(
//		[1,	1,	1,	1,	1,	1,	1,	1,	1],
//		[0,	1,	0,	-1,	0,	1,	-1,	-1,	1],
//		[0,	0,	1,	0,	-1,	1,	1,	-1,	-1],
//		[0,	0,	0,	0,	0,	1,	-1,	1,	-1],
//		[0,	1,	0,	1,	0,	1,	1,	1,	1],
//		[0,	0,	1,	0,	1,	1,	1,	1,	1],
//		[0,	0,	0,	0,	0,	1,	1,	-1,	-1],
//		[0,	0,	0,	0,	0,	1,	-1,	-1,	1],
//		[0,	0,	0,	0,	0,	1,	1,	1,	1]
//	)
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;	A(0,5) = no1;	A(0,6) = no1;	A(0,7) = no1;	A(0,8) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) =-no1;	A(1,4) = 0;	A(1,5) = no1;	A(1,6) =-no1;	A(1,7) =-no1;	A(1,8) = no1;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) =-no1;	A(2,5) = no1;	A(2,6) = no1;	A(2,7) =-no1;	A(2,8) =-no1;
		A(3,0) = 0;	A(3,1) = 0;	A(3,2) = 0;	A(3,3) = 0;	A(3,4) = 0;	A(3,5) = no1;	A(3,6) =-no1;	A(3,7) = no1;	A(3,8) =-no1;
		A(4,0) = 0;	A(4,1) = no1;	A(4,2) = 0;	A(4,3) = no1;	A(4,4) = 0;	A(4,5) = no1;	A(4,6) = no1;	A(4,7) = no1;	A(4,8) = no1;
		A(5,0) = 0;	A(5,1) = 0;	A(5,2) = no1;	A(5,3) = 0;	A(5,4) = no1;	A(5,5) = no1;	A(5,6) = no1;	A(5,7) = no1;	A(5,8) = no1;
		A(6,0) = 0;	A(6,1) = 0;	A(6,2) = 0;	A(6,3) = 0;	A(6,4) = 0;	A(6,5) = no1;	A(6,6) = no1;	A(6,7) =-no1;	A(6,8) =-no1;
		A(7,0) = 0;	A(7,1) = 0;	A(7,2) = 0;	A(7,3) = 0;	A(7,4) = 0;	A(7,5) = no1;	A(7,6) =-no1;	A(7,7) =-no1;	A(7,8) = no1;
		A(8,0) = 0;	A(8,1) = 0;	A(8,2) = 0;	A(8,3) = 0;	A(8,4) = 0;	A(8,5) = no1;	A(8,6) = no1;	A(8,7) = no1;	A(8,8) = no1;
		break;
	case 15:setM_MRT1_15(A);
		break;
	case 19:setM_MRT1_19(A);
		break;
	case 27:setM_MRT1_27(A);
		break;
	default:
		printf("Error: matrix M not implemented for Q%d\n",Q);
	}
}

void LBM::setM_MRT2(matrix &A, int Q)
{
	switch (Q)
	{
	case 3:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = -no1;
		A(2,0) = 0;	A(2,1) = no1;	A(2,2) = no1;
		break;
	case 5:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) =-no1;	A(1,4) = 0;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) =-no1;
		A(3,0) = 0;	A(3,1) = no1;	A(3,2) = no1;	A(3,3) = no1;	A(3,4) = no1;
		A(4,0) = 0;	A(4,1) = no1;	A(4,2) =-no1;	A(4,3) = no1;	A(4,4) =-no1;
		break;
	default:
		printf("Error: matrix M not implemented for Q%d and MRT1\n",Q);
	}
}


void LBM::setC_3(matrix &C)
{
	int maxima_matrix[1][3] = {
		{0,	1,	-1}
	};
	for (int i=0;i<1;i++)
	for (int j=0;j<3;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_5(matrix &C)
{
	int maxima_matrix[2][5] = {
		{0,	1,	0,	-1,	0},
		{0,	0,	1,	0,	-1}
	};
	for (int i=0;i<2;i++)
	for (int j=0;j<5;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_9(matrix &C)
{
	int maxima_matrix[2][9] = {
		{0,	1,	0,	-1,	0,	1,	-1,	-1,	1},
		{0,	0,	1,	0,	-1,	1,	1,	-1,	-1}
	};
	for (int i=0;i<2;i++)
	for (int j=0;j<9;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_7(matrix &C)
{
	int maxima_matrix[3][7] = {
		{0,	1,	0,	0,	-1,	0,	0},
		{0,	0,	1,	0,	0,	-1,	0},
		{0,	0,	0,	1,	0,	0,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<7;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_15(matrix &C)
{
	int maxima_matrix[3][15] = {
		{0,	1,	0,	0,	-1,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<15;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_19(matrix &C)
{
	int maxima_matrix[3][19] = {
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<19;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_27(matrix &C)
{
	int maxima_matrix[3][27] = {
	{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
	{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
	{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<27;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC(matrix &C, int Q)
{
	switch (Q)
	{
	case 3:	setC_3(C);
		break;
	case 5:	setC_5(C);
		break;
	case 7:	setC_7(C);
		break;
	case 9:	setC_9(C);
		break;
	case 15:setC_15(C);
		break;
	case 19:setC_19(C);
		break;
	case 27:setC_27(C);
		break;
	default:
		printf("Error: matrix C not implemented for Q%d\n",Q);
	}
}

void LBM::Feq(matrix &F, ex u, ex v, ex w, ex &sym_cs, int Q, int order)
{
	int islin = (order>=1) ? 1 : 0;
	int isquad = (order>=2) ? 1 : 0;
	switch (Q)
	{
		case 3:
			F(0,0)=(1-sym_cs*sym_cs - isquad*(u*u));
			F(1,0)=(sym_cs*sym_cs + islin*u + isquad*u*u)/2;
			F(2,0)=(sym_cs*sym_cs - islin*u + isquad*u*u)/2;
		break;
		case 5:
			F(0,0)=(1-2*sym_cs*sym_cs - isquad*(u*u + v*v));
			F(1,0)=(sym_cs*sym_cs + islin*u + isquad*u*u)/2;
			F(2,0)=(sym_cs*sym_cs + islin*v + isquad*v*v)/2;
			F(3,0)=(sym_cs*sym_cs - islin*u + isquad*u*u)/2;
			F(4,0)=(sym_cs*sym_cs - islin*v + isquad*v*v)/2;
		break;
		case 7:
			F(0,0)=(1-3*sym_cs*sym_cs - isquad*(u*u + v*v + w*w));
			F(1,0)=(sym_cs*sym_cs + islin*u + isquad*u*u)/2;
			F(2,0)=(sym_cs*sym_cs + islin*v + isquad*v*v)/2;
			F(3,0)=(sym_cs*sym_cs + islin*w + isquad*w*w)/2;
			F(4,0)=(sym_cs*sym_cs - islin*u + isquad*u*u)/2;
			F(5,0)=(sym_cs*sym_cs - islin*v + isquad*v*v)/2;
			F(6,0)=(sym_cs*sym_cs - islin*w + isquad*w*w)/2;
		break;
		case 9:// export from maxima
			F(0,0)= u*u*v*v+sym_cs*sym_cs*v*v-v*v+sym_cs*sym_cs*u*u-u*u+sym_cs*sym_cs*sym_cs*sym_cs-2*sym_cs*sym_cs+1;
			F(1,0)= (-(u*u*v*v)/2)-(u*v*v)/2-(sym_cs*sym_cs*v*v)/2-(sym_cs*sym_cs*u*u)/2+u*u/2-(sym_cs*sym_cs*u)/2+u/2-sym_cs*sym_cs*sym_cs*sym_cs/2+sym_cs*sym_cs/2 ;
			F(2,0)= (-(u*u*v*v)/2)-(sym_cs*sym_cs*v*v)/2+v*v/2-(u*u*v)/2-(sym_cs*sym_cs*v)/2+v/2-(sym_cs*sym_cs*u*u)/2-sym_cs*sym_cs*sym_cs*sym_cs/2+sym_cs*sym_cs/2 ;
			F(3,0)= (-(u*u*v*v)/2)+(u*v*v)/2-(sym_cs*sym_cs*v*v)/2-(sym_cs*sym_cs*u*u)/2+u*u/2+(sym_cs*sym_cs*u)/2-u/2-sym_cs*sym_cs*sym_cs*sym_cs/2+sym_cs*sym_cs/2 ;
			F(4,0)= (-(u*u*v*v)/2)-(sym_cs*sym_cs*v*v)/2+v*v/2+(u*u*v)/2+(sym_cs*sym_cs*v)/2-v/2-(sym_cs*sym_cs*u*u)/2-sym_cs*sym_cs*sym_cs*sym_cs/2+sym_cs*sym_cs/2 ;
			F(5,0)= (u*u*v*v)/4+(u*v*v)/4+(sym_cs*sym_cs*v*v)/4+(u*u*v)/4+(u*v)/4+(sym_cs*sym_cs*v)/4+(sym_cs*sym_cs*u*u)/4+(sym_cs*sym_cs*u)/4+sym_cs*sym_cs*sym_cs*sym_cs/4 ;
			F(6,0)= (u*u*v*v)/4-(u*v*v)/4+(sym_cs*sym_cs*v*v)/4+(u*u*v)/4-(u*v)/4+(sym_cs*sym_cs*v)/4+(sym_cs*sym_cs*u*u)/4-(sym_cs*sym_cs*u)/4+sym_cs*sym_cs*sym_cs*sym_cs/4 ;
			F(7,0)= (u*u*v*v)/4-(u*v*v)/4+(sym_cs*sym_cs*v*v)/4-(u*u*v)/4+(u*v)/4-(sym_cs*sym_cs*v)/4+(sym_cs*sym_cs*u*u)/4-(sym_cs*sym_cs*u)/4+sym_cs*sym_cs*sym_cs*sym_cs/4 ;
			F(8,0)= (u*u*v*v)/4+(u*v*v)/4+(sym_cs*sym_cs*v*v)/4-(u*u*v)/4-(u*v)/4-(sym_cs*sym_cs*v)/4+(sym_cs*sym_cs*u*u)/4+(sym_cs*sym_cs*u)/4+sym_cs*sym_cs*sym_cs*sym_cs/4 ;
		break;
		case 15:
			F(0,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + 1)*2/9;
			F(1,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + u*u/(2*sym_cs*sym_cs*sym_cs*sym_cs) + u/(sym_cs*sym_cs) + 1)/9;
			F(2,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + v*v/(2*sym_cs*sym_cs*sym_cs*sym_cs) + v/(sym_cs*sym_cs) + 1)/9;
			F(3,0)=(w*w/(2*sym_cs*sym_cs*sym_cs*sym_cs) + (-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + w/(sym_cs*sym_cs) + 1)/9;
			F(4,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + u*u/(2*sym_cs*sym_cs*sym_cs*sym_cs) - u/(sym_cs*sym_cs) + 1)/9;
			F(5,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + v*v/(2*sym_cs*sym_cs*sym_cs*sym_cs) - v/(sym_cs*sym_cs) + 1)/9;
			F(6,0)=(w*w/(2*sym_cs*sym_cs*sym_cs*sym_cs) + (-w*w-v*v-u*u)/(2*sym_cs*sym_cs) - w/(sym_cs*sym_cs) + 1)/9;
			F(7,0)=((w+v+u)*(w+v+u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + (-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (w+v+u)/(sym_cs*sym_cs) + 1)/72;
			F(8,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (-w+v+u)/(sym_cs*sym_cs) + (-w+v+u)*(-w+v+u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + 1)/72;
			F(9,0)=((w-v+u)*(w-v+u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + (-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (w-v+u)/(sym_cs*sym_cs) + 1)/72;
			F(10,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (-w-v+u)/(sym_cs*sym_cs) + (-w-v+u)*(-w-v+u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + 1)/72;
			F(11,0)=((w+v-u)*(w+v-u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + (-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (w+v-u)/(sym_cs*sym_cs) + 1)/72;
			F(12,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (-w+v-u)/(sym_cs*sym_cs) + (-w+v-u)*(-w+v-u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + 1)/72;
			F(13,0)=((w-v-u)*(w-v-u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + (-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (w-v-u)/(sym_cs*sym_cs) + 1)/72;
			F(14,0)=((-w*w-v*v-u*u)/(2*sym_cs*sym_cs) + (-w-v-u)/(sym_cs*sym_cs) + (-w-v-u)*(-w-v-u)/(2*sym_cs*sym_cs*sym_cs*sym_cs) + 1)/72;
		break;
		case 27:
			F(0,0)=-(4*w*w)/9-(4*v*v)/9-(4*u*u)/9+(ex)8/27;
			F(1,0)=-(w*w)/9-(v*v)/9+(2*u*u)/9+(2*u)/9+(ex)2/27;
			F(2,0)=-(w*w)/9+(2*v*v)/9+(2*v)/9-(u*u)/9+(ex)2/27;
			F(3,0)=(2*w*w)/9+(2*w)/9-(v*v)/9-(u*u)/9+(ex)2/27;
			F(4,0)=-(w*w)/9-(v*v)/9+(2*u*u)/9-(2*u)/9+(ex)2/27;
			F(5,0)=-(w*w)/9+(2*v*v)/9-(2*v)/9-(u*u)/9+(ex)2/27;
			F(6,0)=(2*w*w)/9-(2*w)/9-(v*v)/9-(u*u)/9+(ex)2/27;
			F(7,0)=(w*w)/18+(v*w)/6+(w)/18+(v*v)/18+(v)/18-(u*u)/36+(ex)1/54;
			F(8,0)=(w*w)/18-(v*w)/6-(w)/18+(v*v)/18+(v)/18-(u*u)/36+(ex)1/54;
			F(9,0)=(w*w)/18-(v*w)/6+(w)/18+(v*v)/18-(v)/18-(u*u)/36+(ex)1/54;
			F(10,0)=(w*w)/18+(v*w)/6-(w)/18+(v*v)/18-(v)/18-(u*u)/36+(ex)1/54;
			F(11,0)=-(w*w)/36+(v*v)/18+(u*v)/6+(v)/18+(u*u)/18+(u)/18+(ex)1/54;
			F(12,0)=-(w*w)/36+(v*v)/18-(u*v)/6-(v)/18+(u*u)/18+(u)/18+(ex)1/54;
			F(13,0)=-(w*w)/36+(v*v)/18-(u*v)/6+(v)/18+(u*u)/18-(u)/18+(ex)1/54;
			F(14,0)=-(w*w)/36+(v*v)/18+(u*v)/6-(v)/18+(u*u)/18-(u)/18+(ex)1/54;
			F(15,0)=(w*w)/18+(u*w)/6+(w)/18-(v*v)/36+(u*u)/18+(u)/18+(ex)1/54;
			F(16,0)=(w*w)/18-(u*w)/6-(w)/18-(v*v)/36+(u*u)/18+(u)/18+(ex)1/54;
			F(17,0)=(w*w)/18-(u*w)/6+(w)/18-(v*v)/36+(u*u)/18-(u)/18+(ex)1/54;
			F(18,0)=(w*w)/18+(u*w)/6-(w)/18-(v*v)/36+(u*u)/18-(u)/18+(ex)1/54;
			F(19,0)=(w*w)/72+(v*w)/24+(u*w)/24+(w)/72+(v*v)/72+(u*v)/24+(v)/72+(u*u)/72+(u)/72+(ex)1/216;
			F(20,0)=(w*w)/72-(v*w)/24-(u*w)/24-(w)/72+(v*v)/72+(u*v)/24+(v)/72+(u*u)/72+(u)/72+(ex)1/216;
			F(21,0)=(w*w)/72-(v*w)/24+(u*w)/24+(w)/72+(v*v)/72-(u*v)/24-(v)/72+(u*u)/72+(u)/72+(ex)1/216;
			F(22,0)=(w*w)/72+(v*w)/24-(u*w)/24-(w)/72+(v*v)/72-(u*v)/24-(v)/72+(u*u)/72+(u)/72+(ex)1/216;
			F(23,0)=(w*w)/72+(v*w)/24-(u*w)/24+(w)/72+(v*v)/72-(u*v)/24+(v)/72+(u*u)/72-(u)/72+(ex)1/216;
			F(24,0)=(w*w)/72-(v*w)/24+(u*w)/24-(w)/72+(v*v)/72-(u*v)/24+(v)/72+(u*u)/72-(u)/72+(ex)1/216;
			F(25,0)=(w*w)/72-(v*w)/24-(u*w)/24+(w)/72+(v*v)/72+(u*v)/24-(v)/72+(u*u)/72-(u)/72+(ex)1/216;
			F(26,0)=(w*w)/72+(v*w)/24+(u*w)/24-(w)/72+(v*v)/72+(u*v)/24-(v)/72+(u*u)/72-(u)/72+(ex)1/216;
		break;
		default: printf("warning: Feq for Q=%d is not implemented yet!\n",Q);
	}
}


bool LBM::moment_symbol_number(char *desc, int q, int Q)
{
	sprintf(desc,"%d",q+1);
	return true;
}

void LBM::moment_symbol(char *desc, int q, int Q)
{
	sprintf(desc,"\\mu");
}
